@@ -20,6 +20,7 @@
 #include "spice-client.h"
 #include "spice-common.h"
 #include "spice-channel-priv.h"
+#include "common/recorder.h"
 
 #include "channel-display-priv.h"
 
@@ -29,8 +30,14 @@
 #include <gst/video/gstvideometa.h>
 
 
+typedef struct SpiceGstFrame SpiceGstFrame;
+
 /* GStreamer decoder implementation */
 
+#if GST_CHECK_VERSION(1,14,0)
+static GstStaticCaps stream_reference = GST_STATIC_CAPS("timestamp/spice-stream");
+#endif
+
 typedef struct SpiceGstDecoder {
     VideoDecoder base;
 
@@ -41,16 +48,15 @@ typedef struct SpiceGstDecoder {
     GstElement *pipeline;
     GstClock *clock;
 
-    guintptr win_handle;
-
     /* ---------- Decoding and display queues ---------- */
 
     uint32_t last_mm_time;
 
     GMutex queues_mutex;
     GQueue *decoding_queue;
-    GQueue *display_queue;
+    SpiceGstFrame *display_frame;
     guint timer_id;
+    guint pending_samples;
 } SpiceGstDecoder;
 
 #define VALID_VIDEO_CODEC_TYPE(codec) \
@@ -76,27 +82,37 @@ typedef enum {
 
 /* ---------- SpiceGstFrame ---------- */
 
-typedef struct SpiceGstFrame {
+struct SpiceGstFrame {
     GstClockTime timestamp;
-    SpiceFrame *frame;
-    GstSample *sample;
-} SpiceGstFrame;
+    GstBuffer *encoded_buffer;
+    SpiceFrame *encoded_frame;
+    GstSample *decoded_sample;
+};
 
 static SpiceGstFrame *create_gst_frame(GstBuffer *buffer, SpiceFrame *frame)
 {
     SpiceGstFrame *gstframe = g_new(SpiceGstFrame, 1);
+
     gstframe->timestamp = GST_BUFFER_PTS(buffer);
-    gstframe->frame = frame;
-    gstframe->sample = NULL;
+#if GST_CHECK_VERSION(1,14,0)
+    GstReferenceTimestampMeta *time_meta;
+
+    time_meta = gst_buffer_get_reference_timestamp_meta(buffer, gst_static_caps_get(&stream_reference));
+    if (time_meta) {
+        gstframe->timestamp = time_meta->timestamp;
+    }
+#endif
+    gstframe->encoded_buffer = gst_buffer_ref(buffer);
+    gstframe->encoded_frame = frame;
+    gstframe->decoded_sample = NULL;
     return gstframe;
 }
 
 static void free_gst_frame(SpiceGstFrame *gstframe)
 {
-    gstframe->frame->free(gstframe->frame);
-    if (gstframe->sample) {
-        gst_sample_unref(gstframe->sample);
-    }
+    gst_buffer_unref(gstframe->encoded_buffer);
+    // encoded_frame was owned by encoded_buffer, don't release it
+    g_clear_pointer(&gstframe->decoded_sample, gst_sample_unref);
     g_free(gstframe);
 }
 
@@ -104,6 +120,10 @@ static void free_gst_frame(SpiceGstFrame *gstframe)
 /* ---------- GStreamer pipeline ---------- */
 
 static void schedule_frame(SpiceGstDecoder *decoder);
+static void fetch_pending_sample(SpiceGstDecoder *decoder);
+static SpiceGstFrame *get_decoded_frame(SpiceGstDecoder *decoder, GstBuffer *buffer);
+
+RECORDER(frames_stats, 64, "Frames statistics");
 
 static int spice_gst_buffer_get_stride(GstBuffer *buffer)
 {
@@ -124,17 +144,18 @@ static gboolean display_frame(gpointer video_decoder)
 
     g_mutex_lock(&decoder->queues_mutex);
     decoder->timer_id = 0;
-    gstframe = g_queue_pop_head(decoder->display_queue);
+    gstframe = decoder->display_frame;
+    decoder->display_frame = NULL;
     g_mutex_unlock(&decoder->queues_mutex);
     /* If the queue is empty we don't even need to reschedule */
     g_return_val_if_fail(gstframe, G_SOURCE_REMOVE);
 
-    if (!gstframe->sample) {
+    if (!gstframe->decoded_sample) {
         spice_warning("got a frame without a sample!");
         goto error;
     }
 
-    caps = gst_sample_get_caps(gstframe->sample);
+    caps = gst_sample_get_caps(gstframe->decoded_sample);
     if (!caps) {
         spice_warning("GStreamer error: could not get the caps of the sample");
         goto error;
@@ -147,13 +168,13 @@ static gboolean display_frame(gpointer video_decoder)
         goto error;
     }
 
-    buffer = gst_sample_get_buffer(gstframe->sample);
+    buffer = gst_sample_get_buffer(gstframe->decoded_sample);
     if (!gst_buffer_map(buffer, &mapinfo, GST_MAP_READ)) {
         spice_warning("GStreamer error: could not map the buffer");
         goto error;
     }
 
-    stream_display_frame(decoder->base.stream, gstframe->frame,
+    stream_display_frame(decoder->base.stream, gstframe->encoded_frame,
                          width, height, spice_gst_buffer_get_stride(buffer), mapinfo.data);
     gst_buffer_unmap(buffer, &mapinfo);
 
@@ -170,25 +191,29 @@ static void schedule_frame(SpiceGstDecoder *decoder)
     g_mutex_lock(&decoder->queues_mutex);
 
     while (!decoder->timer_id) {
-        SpiceGstFrame *gstframe = g_queue_peek_head(decoder->display_queue);
+        while (decoder->display_frame == NULL && decoder->pending_samples) {
+            fetch_pending_sample(decoder);
+        }
+
+        SpiceGstFrame *gstframe = decoder->display_frame;
         if (!gstframe) {
             break;
         }
 
-        if (spice_mmtime_diff(now, gstframe->frame->mm_time) < 0) {
-            decoder->timer_id = g_timeout_add(gstframe->frame->mm_time - now,
+        if (spice_mmtime_diff(now, gstframe->encoded_frame->mm_time) < 0) {
+            decoder->timer_id = g_timeout_add(gstframe->encoded_frame->mm_time - now,
                                               display_frame, decoder);
-        } else if (g_queue_get_length(decoder->display_queue) == 1) {
+        } else if (decoder->display_frame && !decoder->pending_samples) {
             /* Still attempt to display the least out of date frame so the
              * video is not completely frozen for an extended period of time.
              */
             decoder->timer_id = g_timeout_add(0, display_frame, decoder);
         } else {
             SPICE_DEBUG("%s: rendering too late by %u ms (ts: %u, mmtime: %u), dropping",
-                        __FUNCTION__, now - gstframe->frame->mm_time,
-                        gstframe->frame->mm_time, now);
+                        __FUNCTION__, now - gstframe->encoded_frame->mm_time,
+                        gstframe->encoded_frame->mm_time, now);
             stream_dropped_frame_on_playback(decoder->base.stream);
-            g_queue_pop_head(decoder->display_queue);
+            decoder->display_frame = NULL;
             free_gst_frame(gstframe);
         }
     }
@@ -196,6 +221,100 @@ static void schedule_frame(SpiceGstDecoder *decoder)
     g_mutex_unlock(&decoder->queues_mutex);
 }
 
+/* Get the decoded frame relative to buffer or NULL if not found.
+ * Dequeue the frame from decoding_queue and return it, caller
+ * is responsible to free the pointer.
+ * queues_mutex must be held.
+ */
+static SpiceGstFrame *get_decoded_frame(SpiceGstDecoder *decoder, GstBuffer *buffer)
+{
+    GstClockTime buffer_ts = GST_BUFFER_PTS(buffer);
+#if GST_CHECK_VERSION(1,14,0)
+    GstReferenceTimestampMeta *time_meta;
+
+    time_meta = gst_buffer_get_reference_timestamp_meta(buffer, gst_static_caps_get(&stream_reference));
+    if (time_meta) {
+        buffer_ts = time_meta->timestamp;
+    }
+#endif
+
+    /* Gstreamer sometimes returns the same buffer twice
+     * or buffers that have a modified, and thus unrecognizable, PTS.
+     * Blindly removing frames from the decoding_queue until we find a
+     * match would only empty the queue, resulting in later buffers not
+     * finding a match either, etc. So check the buffer has a matching
+     * frame first.
+     */
+    SpiceGstFrame *gstframe = NULL;
+    GList *l = g_queue_peek_head_link(decoder->decoding_queue);
+    while (l) {
+        gstframe = l->data;
+        if (gstframe->timestamp == buffer_ts) {
+            break;
+        }
+        gstframe = NULL;
+        l = l->next;
+    }
+
+    if (gstframe != NULL) {
+        /* Now that we know there is a match, remove it and the older
+         * frames from the decoding queue */
+        SpiceGstFrame *late_frame;
+        guint num_frames_dropped = 0;
+
+        /* The GStreamer pipeline dropped the corresponding buffer. */
+        while ((late_frame = g_queue_pop_head(decoder->decoding_queue)) != gstframe) {
+            num_frames_dropped++;
+            free_gst_frame(late_frame);
+        }
+
+        if (num_frames_dropped != 0) {
+            SPICE_DEBUG("the GStreamer pipeline dropped %u frames", num_frames_dropped);
+        }
+
+        const SpiceFrame *frame = gstframe->encoded_frame;
+        int64_t duration = g_get_monotonic_time() - frame->creation_time;
+        record(frames_stats,
+               "frame mm_time %u size %u creation time %" PRId64
+               " decoded time %" PRId64 " queue %u",
+               frame->mm_time, frame->size, frame->creation_time,
+               duration, decoder->decoding_queue->length);
+    }
+    return gstframe;
+}
+
+static void fetch_pending_sample(SpiceGstDecoder *decoder)
+{
+    GstSample *sample = gst_app_sink_pull_sample(decoder->appsink);
+    if (sample) {
+        // account for the fetched sample
+        decoder->pending_samples--;
+
+        GstBuffer *buffer = gst_sample_get_buffer(sample);
+
+        /* gst_app_sink_pull_sample() sometimes returns the same buffer twice
+         * or buffers that have a modified, and thus unrecognizable, PTS.
+         * Blindly removing frames from the decoding_queue until we find a
+         * match would only empty the queue, resulting in later buffers not
+         * finding a match either, etc. So check the buffer has a matching
+         * frame first.
+         */
+        SpiceGstFrame *gstframe = get_decoded_frame(decoder, buffer);
+        if (gstframe) {
+            /* The frame is now ready for display */
+            gstframe->decoded_sample = sample;
+            decoder->display_frame = gstframe;
+        } else {
+            spice_warning("got an unexpected decoded buffer!");
+            gst_sample_unref(sample);
+        }
+    } else {
+        // no more samples to get, possibly some sample was dropped
+        decoder->pending_samples = 0;
+        spice_warning("GStreamer error: could not pull sample");
+    }
+}
+
 /* GStreamer thread
  *
  * We cannot use GStreamer's signals because they are not always run in
@@ -208,54 +327,16 @@ static GstFlowReturn new_sample(GstAppSink *gstappsink, gpointer video_decoder)
 {
     SpiceGstDecoder *decoder = video_decoder;
 
-    GstSample *sample = gst_app_sink_pull_sample(decoder->appsink);
-    if (sample) {
-        GstBuffer *buffer = gst_sample_get_buffer(sample);
-        g_mutex_lock(&decoder->queues_mutex);
-
-        /* gst_app_sink_pull_sample() sometimes returns the same buffer twice
-         * or buffers that have a modified, and thus unrecognizable, PTS.
-         * Blindly removing frames from the decoding_queue until we find a
-         * match would only empty the queue, resulting in later buffers not
-         * finding a match either, etc. So check the buffer has a matching
-         * frame first.
-         */
-        SpiceGstFrame *gstframe;
-        GList *l = g_queue_peek_head_link(decoder->decoding_queue);
-        while (l) {
-            gstframe = l->data;
-            if (gstframe->timestamp == GST_BUFFER_PTS(buffer)) {
-                /* The frame is now ready for display */
-                gstframe->sample = sample;
-                g_queue_push_tail(decoder->display_queue, gstframe);
-
-                /* Now that we know there is a match, remove it and the older
-                 * frames from the decoding queue.
-                 */
-                while ((gstframe = g_queue_pop_head(decoder->decoding_queue))) {
-                    if (gstframe->timestamp == GST_BUFFER_PTS(buffer)) {
-                        break;
-                    }
-                    /* The GStreamer pipeline dropped the corresponding
-                     * buffer.
-                     */
-                    SPICE_DEBUG("the GStreamer pipeline dropped a frame");
-                    free_gst_frame(gstframe);
-                }
-                break;
-            }
-            l = l->next;
-        }
-        if (!l) {
-            spice_warning("got an unexpected decoded buffer!");
-            gst_sample_unref(sample);
-        }
-
+    g_mutex_lock(&decoder->queues_mutex);
+    decoder->pending_samples++;
+    if (decoder->timer_id && decoder->display_frame) {
         g_mutex_unlock(&decoder->queues_mutex);
-        schedule_frame(decoder);
-    } else {
-        spice_warning("GStreamer error: could not pull sample");
+        return GST_FLOW_OK;
     }
+    g_mutex_unlock(&decoder->queues_mutex);
+
+    schedule_frame(decoder);
+
     return GST_FLOW_OK;
 }
 
@@ -302,27 +383,12 @@ static gboolean handle_pipeline_message(GstBus *bus, GstMessage *msg, gpointer v
                                           gst_opts[decoder->base.codec_type].name);
         GST_DEBUG_BIN_TO_DOT_FILE(GST_BIN(decoder->pipeline),
                                   GST_DEBUG_GRAPH_SHOW_ALL
-#if GST_CHECK_VERSION(1,5,1)
                                     | GST_DEBUG_GRAPH_SHOW_FULL_PARAMS
-#endif
                                     | GST_DEBUG_GRAPH_SHOW_STATES,
                                     filename);
         g_free(filename);
         break;
     }
-    case GST_MESSAGE_ELEMENT: {
-        if (gst_is_video_overlay_prepare_window_handle_message(msg)) {
-            GstVideoOverlay *overlay;
-
-            SPICE_DEBUG("prepare-window-handle msg received (handle: %" PRIuPTR ")", decoder->win_handle);
-            if (decoder->win_handle != 0) {
-                overlay = GST_VIDEO_OVERLAY(GST_MESSAGE_SRC(msg));
-                gst_video_overlay_set_window_handle(overlay, decoder->win_handle);
-                gst_video_overlay_handle_events(overlay, false);
-            }
-        }
-        break;
-    }
     default:
         /* not being handled */
         break;
@@ -330,7 +396,6 @@ static gboolean handle_pipeline_message(GstBus *bus, GstMessage *msg, gpointer v
     return TRUE;
 }
 
-#if GST_CHECK_VERSION(1,9,0)
 static void app_source_setup(GstElement *pipeline G_GNUC_UNUSED,
                              GstElement *source,
                              SpiceGstDecoder *decoder)
@@ -355,12 +420,41 @@ static void app_source_setup(GstElement *pipeline G_GNUC_UNUSED,
     gst_caps_unref(caps);
     decoder->appsrc = GST_APP_SRC(gst_object_ref(source));
 }
-#endif
+
+static GstPadProbeReturn
+sink_event_probe(GstPad *pad, GstPadProbeInfo *info, gpointer data)
+{
+    SpiceGstDecoder *decoder = (SpiceGstDecoder*)data;
+
+    if (info->type & GST_PAD_PROBE_TYPE_BUFFER) { // Buffer arrived
+        GstBuffer *buffer = GST_PAD_PROBE_INFO_BUFFER(info);
+        g_mutex_lock(&decoder->queues_mutex);
+        SpiceGstFrame *gstframe = get_decoded_frame(decoder, buffer);
+        if (gstframe) {
+            free_gst_frame(gstframe);
+        }
+        g_mutex_unlock(&decoder->queues_mutex);
+    }
+    return GST_PAD_PROBE_OK;
+}
+
+/* This function is called to used to set a probe on the sink */
+static void
+deep_element_added_cb(GstBin *pipeline, GstBin *bin, GstElement *element,
+                      SpiceGstDecoder *decoder)
+{
+    /* attach a probe to the sink in case we don't have appsink (that
+     * is the frames does not go to new_sample */
+    if (GST_IS_BASE_SINK(element) && decoder->appsink == NULL) {
+        GstPad *pad = gst_element_get_static_pad(element, "sink");
+        gst_pad_add_probe(pad, GST_PAD_PROBE_TYPE_BUFFER, sink_event_probe, decoder, NULL);
+        gst_object_unref(pad);
+    }
+}
 
 static gboolean create_pipeline(SpiceGstDecoder *decoder)
 {
     GstBus *bus;
-#if GST_CHECK_VERSION(1,9,0)
     GstElement *playbin, *sink;
     SpiceGstPlayFlags flags;
     GstCaps *caps;
@@ -371,14 +465,11 @@ static gboolean create_pipeline(SpiceGstDecoder *decoder)
         return FALSE;
     }
 
-    /* Will try to get window handle in order to apply the GstVideoOverlay
-     * interface, setting overlay to this window will happen only when
-     * prepare-window-handle message is received
+    /* Passing the pipeline to widget, try to get window handle and
+     * set the GstVideoOverlay interface, setting overlay to the window
+     * will happen only when prepare-window-handle message is received
      */
-    decoder->win_handle = get_window_handle(decoder->base.stream);
-    SPICE_DEBUG("Creating Gstreamer pipline (handle for overlay %s)\n",
-                decoder->win_handle ? "received" : "not received");
-    if (decoder->win_handle == 0) {
+    if (!hand_pipeline_to_widget(decoder->base.stream, GST_PIPELINE(playbin))) {
         sink = gst_element_factory_make("appsink", "sink");
         if (sink == NULL) {
             spice_warning("error upon creation of 'appsink' element");
@@ -399,8 +490,12 @@ static gboolean create_pipeline(SpiceGstDecoder *decoder)
         decoder->appsink = GST_APP_SINK(sink);
     } else {
         /* handle has received, it means playbin will render directly into
-         * widget using the gstvideoooverlay interface instead of app-sink.
-         * Also avoid using vaapisink if exist since vaapisink could be
+         * widget using the gstvideooverlay interface instead of app-sink.
+         */
+        SPICE_DEBUG("Video is presented using gstreamer's GstVideoOverlay interface");
+
+#if !GST_CHECK_VERSION(1,14,0)
+        /* Avoid using vaapisink if exist since vaapisink could be
          * buggy when it is combined with playbin. changing its rank to
          * none will make playbin to avoid of using it.
          */
@@ -415,8 +510,10 @@ static gboolean create_pipeline(SpiceGstDecoder *decoder)
             gst_plugin_feature_set_rank(vaapisink, GST_RANK_NONE);
             gst_object_unref(vaapisink);
         }
+#endif
     }
 
+    g_signal_connect(playbin, "deep-element-added", G_CALLBACK(deep_element_added_cb), decoder);
     g_signal_connect(playbin, "source-setup", G_CALLBACK(app_source_setup), decoder);
 
     g_object_set(playbin,
@@ -430,40 +527,13 @@ static gboolean create_pipeline(SpiceGstDecoder *decoder)
 
     g_warn_if_fail(decoder->appsrc == NULL);
     decoder->pipeline = playbin;
-#else
-    gchar *desc;
-    GError *err = NULL;
-
-    /* - We schedule the frame display ourselves so set sync=false on appsink
-     *   so the pipeline decodes them as fast as possible. This will also
-     *   minimize the risk of frames getting lost when we rebuild the
-     *   pipeline.
-     * - Set max-bytes=0 on appsrc so it does not drop frames that may be
-     *   needed by those that follow.
-     */
-    desc = g_strdup_printf("appsrc name=src is-live=true format=time max-bytes=0 block=true "
-                           "caps=%s ! %s ! videoconvert ! appsink name=sink "
-                           "caps=video/x-raw,format=BGRx sync=false drop=false",
-                           gst_opts[decoder->base.codec_type].dec_caps,
-                           gst_opts[decoder->base.codec_type].dec_name);
-    SPICE_DEBUG("GStreamer pipeline: %s", desc);
-
-    decoder->pipeline = gst_parse_launch_full(desc, NULL, GST_PARSE_FLAG_FATAL_ERRORS, &err);
-    g_free(desc);
-    if (!decoder->pipeline) {
-        spice_warning("GStreamer error: %s", err->message);
-        g_clear_error(&err);
-        return FALSE;
-    }
-
-    decoder->appsrc = GST_APP_SRC(gst_bin_get_by_name(GST_BIN(decoder->pipeline), "src"));
-    decoder->appsink = GST_APP_SINK(gst_bin_get_by_name(GST_BIN(decoder->pipeline), "sink"));
-#endif
 
     if (decoder->appsink) {
         GstAppSinkCallbacks appsink_cbs = { NULL };
         appsink_cbs.new_sample = new_sample;
         gst_app_sink_set_callbacks(decoder->appsink, &appsink_cbs, decoder, NULL);
+        gst_app_sink_set_max_buffers(decoder->appsink, 2);
+        gst_app_sink_set_drop(decoder->appsink, FALSE);
     }
     bus = gst_pipeline_get_bus(GST_PIPELINE(decoder->pipeline));
     gst_bus_add_watch(bus, handle_pipeline_message, decoder);
@@ -486,6 +556,10 @@ static gboolean create_pipeline(SpiceGstDecoder *decoder)
 static void spice_gst_decoder_reschedule(VideoDecoder *video_decoder)
 {
     SpiceGstDecoder *decoder = (SpiceGstDecoder*)video_decoder;
+
+    if (!decoder->appsink) {
+        return;
+    }
     guint timer_id;
 
     g_mutex_lock(&decoder->queues_mutex);
@@ -517,15 +591,10 @@ static void spice_gst_decoder_destroy(VideoDecoder *video_decoder)
         g_source_remove(decoder->timer_id);
     }
     g_mutex_clear(&decoder->queues_mutex);
-    SpiceGstFrame *gstframe;
-    while ((gstframe = g_queue_pop_head(decoder->decoding_queue))) {
-        free_gst_frame(gstframe);
+    g_queue_free_full(decoder->decoding_queue, (GDestroyNotify)free_gst_frame);
+    if (decoder->display_frame) {
+        free_gst_frame(decoder->display_frame);
     }
-    g_queue_free(decoder->decoding_queue);
-    while ((gstframe = g_queue_pop_head(decoder->display_queue))) {
-        free_gst_frame(gstframe);
-    }
-    g_queue_free(decoder->display_queue);
 
     g_free(decoder);
 
@@ -538,27 +607,39 @@ static void spice_gst_decoder_destroy(VideoDecoder *video_decoder)
 /* spice_gst_decoder_queue_frame() queues the SpiceFrame for decoding and
  * displaying. The steps it goes through are as follows:
  *
- * 1) A SpiceGstFrame is created to keep track of SpiceFrame and some additional
- *    metadata. The SpiceGstFrame is then pushed to the decoding_queue.
- * 2) frame->data, which contains the compressed frame data, is reffed and
- *    wrapped in a GstBuffer which is pushed to the GStreamer pipeline for
- *    decoding.
- * 3) As soon as the GStreamer pipeline no longer needs the compressed frame it
- *    will call frame->unref_data() to free it.
- * 4) Once the decompressed frame is available the GStreamer pipeline calls
- *    new_sample() in the GStreamer thread.
- * 5) new_sample() then matches the decompressed frame to a SpiceGstFrame from
- *    the decoding queue using the GStreamer timestamp information to deal with
- *    dropped frames. The SpiceGstFrame is popped from the decoding_queue.
- * 6) new_sample() then attaches the decompressed frame to the SpiceGstFrame,
- *    pushes it to the display_queue and calls schedule_frame().
- * 7) schedule_frame() then uses gstframe->frame->mm_time to arrange for
- *    display_frame() to be called, in the main thread, at the right time for
- *    the next frame.
- * 8) display_frame() pops the first SpiceGstFrame from the display_queue and
- *    calls stream_display_frame().
- * 9) display_frame() then frees the SpiceGstFrame, which frees the SpiceFrame
- *    and decompressed frame with it.
+ * 1) frame->data, which contains the compressed frame data, is wrapped in a GstBuffer
+ *    (encoded_buffer) which owns the SpiceFrame.
+ * 2) A SpiceGstFrame is created to keep track of SpiceFrame (encoded_frame), and some
+ *    additional metadata. The encoded_buffer is referenced and the SpiceGstFrame is then
+ *    pushed into the decoding_queue.
+ *
+ * If GstVideoOverlay is used (window handle was obtained successfully at the widget):
+ *   3) Decompressed frames will be rendered to widget directly from GStreamer's pipeline
+ *      using some GStreamer sink plugin which implements the GstVideoOverlay interface
+ *      (last step).
+ *   4) As soon as GStreamer's pipeline no longer needs the compressed frame it will
+ *      unref the encoded_buffer
+ *   5) Once a decoded buffer arrives to the sink (notified by probe event) we will pop
+ *      its matching SpiceGstFrame from the decoding_queue and free it using
+ *      free_gst_frame, this will also unref the encoded_buffer which will allow
+ *      GStreamer to call spice_frame_free and free its encoded_frame.
+ *
+ * Otherwise appsink is used:
+ *   3) Once the decompressed frame is available the GStreamer pipeline calls
+ *      new_sample() in the GStreamer thread.
+ *   4) new_sample() then matches the decompressed frame to a SpiceGstFrame from
+ *      the decoding queue using the GStreamer timestamp information to deal with
+ *      dropped frames. The SpiceGstFrame is popped from the decoding_queue.
+ *   5) new_sample() then attaches the decompressed frame to the SpiceGstFrame,
+ *      set into display_frame and calls schedule_frame().
+ *   6) schedule_frame() then uses gstframe->encoded_frame->mm_time to arrange for
+ *      display_frame() to be called, in the main thread, at the right time for
+ *      the next frame.
+ *   7) display_frame() uses SpiceGstFrame from display_frame and calls
+ *      stream_display_frame().
+ *   8) display_frame() then calls to free_gst_frame to free the SpiceGstFrame and
+ *      unref the encoded_buffer which allows GStreamer to call spice_frame_free
+ *      and free its encoded_frame.
  */
 static gboolean spice_gst_decoder_queue_frame(VideoDecoder *video_decoder,
                                               SpiceFrame *frame, int latency)
@@ -567,7 +648,7 @@ static gboolean spice_gst_decoder_queue_frame(VideoDecoder *video_decoder,
 
     if (frame->size == 0) {
         SPICE_DEBUG("got an empty frame buffer!");
-        frame->free(frame);
+        spice_frame_free(frame);
         return TRUE;
     }
 
@@ -585,45 +666,42 @@ static gboolean spice_gst_decoder_queue_frame(VideoDecoder *video_decoder,
          * saves CPU so do it.
          */
         SPICE_DEBUG("dropping a late MJPEG frame");
-        frame->free(frame);
+        spice_frame_free(frame);
         return TRUE;
     }
 
     if (decoder->pipeline == NULL) {
         /* An error occurred, causing the GStreamer pipeline to be freed */
         spice_warning("An error occurred, stopping the video stream");
-        frame->free(frame);
+        spice_frame_free(frame);
         return FALSE;
     }
 
-#if GST_CHECK_VERSION(1,9,0)
     if (decoder->appsrc == NULL) {
         spice_warning("Error: Playbin has not yet initialized the Appsrc element");
         stream_dropped_frame_on_playback(decoder->base.stream);
-        frame->free(frame);
+        spice_frame_free(frame);
         return TRUE;
     }
-#endif
 
-    /* ref() the frame data for the buffer */
-    frame->ref_data(frame->data_opaque);
+    /* frame ownership is moved to the buffer */
     GstBuffer *buffer = gst_buffer_new_wrapped_full(GST_MEMORY_FLAG_PHYSICALLY_CONTIGUOUS,
                                                     frame->data, frame->size, 0, frame->size,
-                                                    frame->data_opaque, frame->unref_data);
+                                                    frame, (GDestroyNotify) spice_frame_free);
 
+    GstClockTime pts = gst_clock_get_time(decoder->clock) - gst_element_get_base_time(decoder->pipeline) + ((uint64_t)MAX(0, latency)) * 1000 * 1000;
     GST_BUFFER_DURATION(buffer) = GST_CLOCK_TIME_NONE;
     GST_BUFFER_DTS(buffer) = GST_CLOCK_TIME_NONE;
-    GST_BUFFER_PTS(buffer) = gst_clock_get_time(decoder->clock) - gst_element_get_base_time(decoder->pipeline) + ((uint64_t)MAX(0, latency)) * 1000 * 1000;
+    GST_BUFFER_PTS(buffer) = pts;
+#if GST_CHECK_VERSION(1,14,0)
+    gst_buffer_add_reference_timestamp_meta(buffer, gst_static_caps_get(&stream_reference),
+                                            pts, GST_CLOCK_TIME_NONE);
+#endif
 
-    if (decoder->appsink != NULL) {
-        SpiceGstFrame *gst_frame = create_gst_frame(buffer, frame);
-        g_mutex_lock(&decoder->queues_mutex);
-        g_queue_push_tail(decoder->decoding_queue, gst_frame);
-        g_mutex_unlock(&decoder->queues_mutex);
-    } else {
-        frame->free(frame);
-        frame = NULL;
-    }
+    SpiceGstFrame *gst_frame = create_gst_frame(buffer, frame);
+    g_mutex_lock(&decoder->queues_mutex);
+    g_queue_push_tail(decoder->decoding_queue, gst_frame);
+    g_mutex_unlock(&decoder->queues_mutex);
 
     if (gst_app_src_push_buffer(decoder->appsrc, buffer) != GST_FLOW_OK) {
         SPICE_DEBUG("GStreamer error: unable to push frame");
@@ -664,7 +742,6 @@ VideoDecoder* create_gstreamer_decoder(int codec_type, display_stream *stream)
         decoder->base.stream = stream;
         g_mutex_init(&decoder->queues_mutex);
         decoder->decoding_queue = g_queue_new();
-        decoder->display_queue = g_queue_new();
 
         if (!create_pipeline(decoder)) {
             decoder->base.destroy((VideoDecoder*)decoder);
@@ -712,7 +789,8 @@ gboolean gstvideo_has_codec(int codec_type)
            GST_ELEMENT_FACTORY_TYPE_MEDIA_IMAGE;
     all_decoders = gst_element_factory_list_get_elements(type, GST_RANK_NONE);
     if (all_decoders == NULL) {
-        spice_warning("No video decoders from GStreamer were found");
+        spice_debug("No video decoders from GStreamer for %s were found",
+                    gst_opts[codec_type].name);
         return FALSE;
     }
 
