@@ -89,6 +89,7 @@ enum {
     SPICE_DISPLAY_MARK,
     SPICE_DISPLAY_GL_DRAW,
     SPICE_DISPLAY_STREAMING_MODE,
+    SPICE_DISPLAY_OVERLAY,
 
     SPICE_DISPLAY_LAST_SIGNAL,
 };
@@ -102,7 +103,7 @@ static void clear_surfaces(SpiceChannel *channel, gboolean keep_primary);
 static void clear_streams(SpiceChannel *channel);
 static display_surface *find_surface(SpiceDisplayChannelPrivate *c, guint32 surface_id);
 static void spice_display_channel_reset(SpiceChannel *channel, gboolean migrating);
-static void spice_display_channel_reset_capabilities(SpiceChannel *channel);
+static void spice_display_channel_set_capabilities(SpiceChannel *channel);
 static void destroy_canvas(display_surface *surface);
 static void display_stream_destroy(gpointer st);
 static void display_session_mm_time_reset_cb(SpiceSession *session, gpointer data);
@@ -191,6 +192,7 @@ static void spice_display_channel_constructed(GObject *object)
                                   G_CALLBACK(display_session_mm_time_reset_cb),
                                   SPICE_CHANNEL(object), 0);
 
+    spice_display_channel_set_capabilities(SPICE_CHANNEL(object));
 
     if (G_OBJECT_CLASS(spice_display_channel_parent_class)->constructed)
         G_OBJECT_CLASS(spice_display_channel_parent_class)->constructed(object);
@@ -266,7 +268,6 @@ static void spice_display_channel_class_init(SpiceDisplayChannelClass *klass)
 
     channel_class->channel_up   = spice_display_channel_up;
     channel_class->channel_reset = spice_display_channel_reset;
-    channel_class->channel_reset_capabilities = spice_display_channel_reset_capabilities;
 
     g_object_class_install_property
         (gobject_class, PROP_HEIGHT,
@@ -435,7 +436,7 @@ static void spice_display_channel_class_init(SpiceDisplayChannelClass *klass)
      * @width: width
      * @height: height
      *
-     * The #SpiceDisplayChannel::draw signal is emitted when the
+     * The #SpiceDisplayChannel::gl-draw signal is emitted when the
      * rectangular region x/y/w/h of the GL scanout is updated and
      * must be drawn. When the draw is finished, you must call
      * spice_display_gl_draw_done() in order to release the GL
@@ -462,18 +463,43 @@ static void spice_display_channel_class_init(SpiceDisplayChannelClass *klass)
      * The #SpiceDisplayChannel::streaming-mode signal is emitted when
      * spice server is working in streaming mode.
      *
-     * Since 0.35
+     * Since: 0.35
+     *
+     * Deprecated: 0.36: use #SpiceDisplayChannel::gst-video-overlay
+     * instead
      **/
     signals[SPICE_DISPLAY_STREAMING_MODE] =
         g_signal_new("streaming-mode",
                      G_OBJECT_CLASS_TYPE(gobject_class),
-                     0, 0,
-                     NULL, NULL,
-                     g_cclosure_user_marshal_POINTER__BOOLEAN,
+                     G_SIGNAL_DEPRECATED, 0,
+                     NULL, NULL, NULL,
                      G_TYPE_POINTER,
                      1,
                      G_TYPE_BOOLEAN);
 
+    /**
+     * SpiceDisplayChannel::gst-video-overlay:
+     * @display: the #SpiceDisplayChannel that emitted the signal
+     * @pipeline: pointer to GStreamer's pipeline
+     *
+     * The #SpiceDisplayChannel::gst-video-overlay signal is emitted when
+     * pipeline is ready and can be passed to widget to register GStreamer
+     * overlay interface and other GStreamer callbacks.
+     *
+     * Returns: %TRUE if the overlay is being set
+     *
+     * Since: 0.36
+     **/
+    signals[SPICE_DISPLAY_OVERLAY] =
+        g_signal_new("gst-video-overlay",
+                     G_OBJECT_CLASS_TYPE(gobject_class),
+                     0, 0,
+                     NULL, NULL,
+                     g_cclosure_user_marshal_BOOLEAN__POINTER,
+                     G_TYPE_BOOLEAN,
+                     1,
+                     GST_TYPE_PIPELINE);
+
     channel_set_handlers(SPICE_CHANNEL_CLASS(klass));
 }
 
@@ -788,7 +814,10 @@ static void image_put_lossy(SpiceImageCache *cache, uint64_t id,
 static void image_replace_lossy(SpiceImageCache *cache, uint64_t id,
                                 pixman_image_t *surface)
 {
-    image_put(cache, id, surface);
+    SpiceDisplayChannelPrivate *c =
+        SPICE_CONTAINEROF(cache, SpiceDisplayChannelPrivate, image_cache);
+
+    cache_replace_lossy(c->images, id, pixman_image_ref(surface), FALSE);
 }
 
 static pixman_image_t* image_get_lossless(SpiceImageCache *cache, uint64_t id)
@@ -836,30 +865,31 @@ static SpiceImageSurfacesOps image_surfaces_ops = {
     .get = surfaces_get
 };
 
-static void spice_display_channel_reset_capabilities(SpiceChannel *channel)
+static void spice_display_channel_set_capabilities(SpiceChannel *channel)
 {
+    SpiceSession *s = spice_channel_get_session(channel);
     guint i;
 
-    spice_channel_set_capability(SPICE_CHANNEL(channel), SPICE_DISPLAY_CAP_SIZED_STREAM);
-    spice_channel_set_capability(SPICE_CHANNEL(channel), SPICE_DISPLAY_CAP_MONITORS_CONFIG);
-    spice_channel_set_capability(SPICE_CHANNEL(channel), SPICE_DISPLAY_CAP_COMPOSITE);
-    spice_channel_set_capability(SPICE_CHANNEL(channel), SPICE_DISPLAY_CAP_A8_SURFACE);
+    spice_channel_set_capability(channel, SPICE_DISPLAY_CAP_SIZED_STREAM);
+    spice_channel_set_capability(channel, SPICE_DISPLAY_CAP_MONITORS_CONFIG);
+    spice_channel_set_capability(channel, SPICE_DISPLAY_CAP_COMPOSITE);
+    spice_channel_set_capability(channel, SPICE_DISPLAY_CAP_A8_SURFACE);
 #ifdef USE_LZ4
-    spice_channel_set_capability(SPICE_CHANNEL(channel), SPICE_DISPLAY_CAP_LZ4_COMPRESSION);
+    spice_channel_set_capability(channel, SPICE_DISPLAY_CAP_LZ4_COMPRESSION);
 #endif
     if (SPICE_DISPLAY_CHANNEL(channel)->priv->enable_adaptive_streaming) {
-        spice_channel_set_capability(SPICE_CHANNEL(channel), SPICE_DISPLAY_CAP_STREAM_REPORT);
+        spice_channel_set_capability(channel, SPICE_DISPLAY_CAP_STREAM_REPORT);
     }
-#ifdef G_OS_UNIX
-    spice_channel_set_capability(SPICE_CHANNEL(channel), SPICE_DISPLAY_CAP_GL_SCANOUT);
-#endif
-    spice_channel_set_capability(SPICE_CHANNEL(channel), SPICE_DISPLAY_CAP_MULTI_CODEC);
+    if (spice_session_get_gl_scanout_enabled(s)) {
+        spice_channel_set_capability(channel, SPICE_DISPLAY_CAP_GL_SCANOUT);
+    }
+    spice_channel_set_capability(channel, SPICE_DISPLAY_CAP_MULTI_CODEC);
 #ifdef HAVE_BUILTIN_MJPEG
-    spice_channel_set_capability(SPICE_CHANNEL(channel), SPICE_DISPLAY_CAP_CODEC_MJPEG);
+    spice_channel_set_capability(channel, SPICE_DISPLAY_CAP_CODEC_MJPEG);
 #endif
     for (i = 1; i < G_N_ELEMENTS(gst_opts); i++) {
         if (gstvideo_has_codec(i)) {
-            spice_channel_set_capability(SPICE_CHANNEL(channel), gst_opts[i].cap);
+            spice_channel_set_capability(channel, gst_opts[i].cap);
         } else {
             SPICE_DEBUG("GStreamer does not support the %s codec", gst_opts[i].name);
         }
@@ -893,7 +923,6 @@ static void spice_display_channel_init(SpiceDisplayChannel *channel)
     } else {
         c->enable_adaptive_streaming = TRUE;
     }
-    spice_display_channel_reset_capabilities(SPICE_CHANNEL(channel));
 }
 
 /* ------------------------------------------------------------------ */
@@ -1211,7 +1240,7 @@ static void report_invalid_stream(SpiceChannel *channel, uint32_t id)
         /* Send a special stream report (UINT_MAX dropped frames out of zero)
          * to indicate there is no such stream.
          */
-        spice_printerr("notify the server that stream %u does not exist", id);
+        g_warning("notify the server that stream %u does not exist", id);
         memset(&report, 0, sizeof(report));
         report.stream_id = id;
         report.num_frames = 0;
@@ -1263,13 +1292,11 @@ static display_stream *display_stream_create(SpiceChannel *channel,
         break;
 #endif
     default:
-#ifdef HAVE_GSTVIDEO
         st->video_decoder = create_gstreamer_decoder(codec_type, st);
-#endif
         break;
     }
     if (st->video_decoder == NULL) {
-        spice_printerr("could not create a video decoder for codec %u", codec_type);
+        g_warning("could not create a video decoder for codec %u", codec_type);
         g_clear_pointer(&st, display_stream_destroy);
     }
     return st;
@@ -1310,7 +1337,7 @@ static void display_handle_stream_create(SpiceChannel *channel, SpiceMsgIn *in)
                                                op->flags, op->codec_type,
                                                &op->dest, &op->clip);
     if (c->streams[op->id] == NULL) {
-        spice_printerr("could not create the %u video stream", op->id);
+        g_warning("could not create the %u video stream", op->id);
         destroy_stream(channel, op->id);
         report_invalid_stream(channel, op->id);
     }
@@ -1388,13 +1415,16 @@ void stream_display_frame(display_stream *st, SpiceFrame *frame,
     }
 }
 
-guintptr get_window_handle(display_stream *st)
+G_GNUC_INTERNAL
+gboolean hand_pipeline_to_widget(display_stream *st, GstPipeline *pipeline)
 {
-   void* handle = 0;
+    gboolean res = false;
 
-   g_signal_emit(st->channel, signals[SPICE_DISPLAY_STREAMING_MODE], 0,
-                 st->surface->streaming_mode, &handle);
-   return st->surface->streaming_mode ? (guintptr)handle : 0;
+    if (st->surface->streaming_mode) {
+        g_signal_emit(st->channel, signals[SPICE_DISPLAY_OVERLAY], 0,
+                      pipeline, &res);
+    }
+    return res;
 }
 
 /* after a sequence of 3 drops, push a report to the server, even
@@ -1515,6 +1545,133 @@ static void display_session_mm_time_reset_cb(SpiceSession *session, gpointer dat
 
 #define STREAM_PLAYBACK_SYNC_DROP_SEQ_LEN_LIMIT 5
 
+static void display_stream_stats_debug(display_stream *st)
+{
+    guint64 drops_duration_total = 0;
+    guint32 i, num_out_frames;
+    gdouble avg_late_time = 0.0;
+
+    if (st->num_input_frames == 0) {
+        return;
+    }
+
+    num_out_frames = st->num_input_frames - st->arrive_late_count - st->num_drops_on_playback;
+
+    if (st->arrive_late_count != 0) {
+        avg_late_time = st->arrive_late_time / ((double)st->arrive_late_count);
+    }
+
+    CHANNEL_DEBUG(st->channel,
+        "%s: id=%u #in-frames=%u out/in=%.2f "
+        "#drops-on-receive=%u avg-late-time(ms)=%.2f "
+        "#drops-on-playback=%u",
+        __FUNCTION__,
+        st->id,
+        st->num_input_frames,
+        num_out_frames / (double)st->num_input_frames,
+        st->arrive_late_count,
+        avg_late_time,
+        st->num_drops_on_playback);
+
+    if (st->num_drops_seqs) {
+        CHANNEL_DEBUG(st->channel,
+                      "%s: #drops-sequences=%u ==>",
+                      __FUNCTION__,
+                      st->num_drops_seqs);
+    }
+
+    for (i = 0; i < st->num_drops_seqs; i++) {
+        drops_sequence_stats *stats = &g_array_index(st->drops_seqs_stats_arr,
+                                                     drops_sequence_stats,
+                                                     i);
+        drops_duration_total += stats->duration;
+        CHANNEL_DEBUG(st->channel,
+                      "%s: \t len=%u start-ms=%u duration-ms=%u",
+                      __FUNCTION__,
+                      stats->len,
+                      stats->start_mm_time - st->first_frame_mm_time,
+                      stats->duration);
+    }
+
+    if (st->num_drops_seqs) {
+        CHANNEL_DEBUG(st->channel,
+                      "%s: drops-total-duration=%"G_GUINT64_FORMAT" ==>",
+                      __FUNCTION__,
+                      drops_duration_total);
+    }
+}
+
+
+static void display_stream_stats_save(display_stream *st,
+                                      guint32 server_mmtime,
+                                      guint32 client_mmtime)
+{
+    gint32 latency = server_mmtime - client_mmtime;
+
+    if (!st->num_input_frames) {
+        st->first_frame_mm_time = server_mmtime;
+    }
+    st->num_input_frames++;
+
+    if (latency < 0) {
+        CHANNEL_DEBUG(st->channel, "stream data too late by %u ms (ts: %u, mmtime: %u)",
+                      client_mmtime - server_mmtime, server_mmtime, client_mmtime);
+        st->arrive_late_time += client_mmtime - server_mmtime;
+        st->arrive_late_count++;
+
+        /* Late frames are counted as drops in the stats but aren't necessarily dropped - depends
+         * on codec and decoder
+         */
+        if (!st->cur_drops_seq_stats.len) {
+            st->cur_drops_seq_stats.start_mm_time = server_mmtime;
+        }
+        st->cur_drops_seq_stats.len++;
+        st->playback_sync_drops_seq_len++;
+        return;
+    }
+
+    CHANNEL_DEBUG(st->channel, "video latency: %d", latency);
+    if (st->cur_drops_seq_stats.len) {
+        st->cur_drops_seq_stats.duration = server_mmtime -
+                                           st->cur_drops_seq_stats.start_mm_time;
+        g_array_append_val(st->drops_seqs_stats_arr, st->cur_drops_seq_stats);
+        memset(&st->cur_drops_seq_stats, 0, sizeof(st->cur_drops_seq_stats));
+        st->num_drops_seqs++;
+    }
+    st->playback_sync_drops_seq_len = 0;
+}
+
+static SpiceFrame *spice_frame_new(display_stream *st,
+                                   SpiceMsgIn *in,
+                                   guint32 server_mmtime)
+{
+    SpiceFrame *frame;
+    guint8 *data_ptr;
+    const SpiceRect *dest_rect = stream_get_dest(st, in);
+    guint32 data_size = spice_msg_in_frame_data(in, &data_ptr);
+
+    frame = g_new(SpiceFrame, 1);
+    frame->mm_time = server_mmtime;
+    frame->dest = *dest_rect;
+    frame->data = data_ptr;
+    frame->size = data_size;
+    frame->data_opaque = in;
+    spice_msg_in_ref(in);
+    frame->creation_time = g_get_monotonic_time();
+    return frame;
+}
+
+G_GNUC_INTERNAL
+void spice_frame_free(SpiceFrame *frame)
+{
+    if (frame == NULL) {
+        return;
+    }
+
+    spice_msg_in_unref(frame->data_opaque);
+    g_free(frame);
+}
+
 /* coroutine context */
 static void display_handle_stream_data(SpiceChannel *channel, SpiceMsgIn *in)
 {
@@ -1522,7 +1679,7 @@ static void display_handle_stream_data(SpiceChannel *channel, SpiceMsgIn *in)
     SpiceStreamDataHeader *op = spice_msg_in_parsed(in);
     display_stream *st = get_stream_by_id(channel, op->id);
     guint32 mmtime;
-    int32_t latency;
+    int32_t latency, latency_report;
     SpiceFrame *frame;
 
     g_return_if_fail(st != NULL);
@@ -1537,47 +1694,22 @@ static void display_handle_stream_data(SpiceChannel *channel, SpiceMsgIn *in)
         op->multi_media_time = mmtime + 100; /* workaround... */
     }
 
-    if (!st->num_input_frames) {
-        st->first_frame_mm_time = op->multi_media_time;
-    }
-    st->num_input_frames++;
+    latency = latency_report = op->multi_media_time - mmtime;
+    if (latency > 0) {
+        SpiceSession *s = spice_channel_get_session(channel);
 
-    latency = op->multi_media_time - mmtime;
-    if (latency < 0) {
-        CHANNEL_DEBUG(channel, "stream data too late by %u ms (ts: %u, mmtime: %u), dropping",
-                      mmtime - op->multi_media_time, op->multi_media_time, mmtime);
-        st->arrive_late_time += mmtime - op->multi_media_time;
-        st->arrive_late_count++;
-
-        if (!st->cur_drops_seq_stats.len) {
-            st->cur_drops_seq_stats.start_mm_time = op->multi_media_time;
+        if (st->surface->streaming_mode && !spice_session_is_playback_active(s)) {
+            CHANNEL_DEBUG(channel, "video latency: %d, set to 0 since there is no playback", latency);
+            latency = 0;
         }
-        st->cur_drops_seq_stats.len++;
-        st->playback_sync_drops_seq_len++;
-    } else {
-        CHANNEL_DEBUG(channel, "video latency: %d", latency);
-        if (st->cur_drops_seq_stats.len) {
-            st->cur_drops_seq_stats.duration = op->multi_media_time -
-                                               st->cur_drops_seq_stats.start_mm_time;
-            g_array_append_val(st->drops_seqs_stats_arr, st->cur_drops_seq_stats);
-            memset(&st->cur_drops_seq_stats, 0, sizeof(st->cur_drops_seq_stats));
-            st->num_drops_seqs++;
-        }
-        st->playback_sync_drops_seq_len = 0;
     }
+    display_stream_stats_save(st, op->multi_media_time, mmtime);
 
     /* Let the video decoder queue the frames so it can optimize their
      * decoding and best decide if/when to drop them when they are late,
      * taking into account the impact on later frames.
      */
-    frame = g_new(SpiceFrame, 1);
-    frame->mm_time = op->multi_media_time;
-    frame->dest = *stream_get_dest(st, in);
-    frame->size = spice_msg_in_frame_data(in, &frame->data);
-    frame->data_opaque = in;
-    frame->ref_data = (void*)spice_msg_in_ref;
-    frame->unref_data = (void*)spice_msg_in_unref;
-    frame->free = (void*)g_free;
+    frame = spice_frame_new(st, in, op->multi_media_time);
     if (!st->video_decoder->queue_frame(st->video_decoder, frame, latency)) {
         destroy_stream(channel, op->id);
         report_invalid_stream(channel, op->id);
@@ -1586,7 +1718,7 @@ static void display_handle_stream_data(SpiceChannel *channel, SpiceMsgIn *in)
 
     if (c->enable_adaptive_streaming) {
         display_update_stream_report(SPICE_DISPLAY_CHANNEL(channel), op->id,
-                                     op->multi_media_time, latency);
+                                     op->multi_media_time, latency_report);
         if (st->playback_sync_drops_seq_len >= STREAM_PLAYBACK_SYNC_DROP_SEQ_LEN_LIMIT) {
             spice_session_sync_playback_latency(spice_channel_get_session(channel));
             st->playback_sync_drops_seq_len = 0;
@@ -1608,39 +1740,9 @@ static void display_handle_stream_clip(SpiceChannel *channel, SpiceMsgIn *in)
 
 static void display_stream_destroy(gpointer st_pointer)
 {
-    int i;
     display_stream *st = st_pointer;
 
-    if (st->num_input_frames > 0) {
-        guint64 drops_duration_total = 0;
-        guint32 num_out_frames = st->num_input_frames - st->arrive_late_count - st->num_drops_on_playback;
-        CHANNEL_DEBUG(st->channel, "%s: id=%u #in-frames=%u out/in=%.2f "
-            "#drops-on-receive=%u avg-late-time(ms)=%.2f "
-            "#drops-on-playback=%u", __FUNCTION__,
-            st->id,
-            st->num_input_frames,
-            num_out_frames / (double)st->num_input_frames,
-            st->arrive_late_count,
-            st->arrive_late_count ? st->arrive_late_time / ((double)st->arrive_late_count): 0,
-            st->num_drops_on_playback);
-        if (st->num_drops_seqs) {
-            CHANNEL_DEBUG(st->channel, "%s: #drops-sequences=%u ==>", __FUNCTION__, st->num_drops_seqs);
-        }
-        for (i = 0; i < st->num_drops_seqs; i++) {
-            drops_sequence_stats *stats = &g_array_index(st->drops_seqs_stats_arr,
-                                                         drops_sequence_stats,
-                                                         i);
-            drops_duration_total += stats->duration;
-            CHANNEL_DEBUG(st->channel, "%s: \t len=%u start-ms=%u duration-ms=%u", __FUNCTION__,
-                          stats->len,
-                          stats->start_mm_time - st->first_frame_mm_time,
-                          stats->duration);
-        }
-        if (st->num_drops_seqs) {
-            CHANNEL_DEBUG(st->channel, "%s: drops-total-duration=%"G_GUINT64_FORMAT" ==>", __FUNCTION__, drops_duration_total);
-        }
-    }
-
+    display_stream_stats_debug(st);
     g_array_free(st->drops_seqs_stats_arr, TRUE);
 
     if (st->video_decoder) {
@@ -1893,9 +1995,9 @@ static void display_handle_monitors_config(SpiceChannel *channel, SpiceMsgIn *in
         SpiceDisplayMonitorConfig *mc = &g_array_index(c->monitors, SpiceDisplayMonitorConfig, i);
         SpiceHead *head = &config->heads[i];
         CHANNEL_DEBUG(channel, "monitor id: %u, surface id: %u, +%u+%u-%ux%u",
-                    head->id, head->surface_id,
+                    head->monitor_id, head->surface_id,
                     head->x, head->y, head->width, head->height);
-        mc->id = head->id;
+        mc->id = head->monitor_id;
         mc->surface_id = head->surface_id;
         mc->x = head->x;
         mc->y = head->y;
