@@ -33,7 +33,6 @@
 #include "channel-usbredir-priv.h"
 #include "usb-device-manager-priv.h"
 #include "usbutil.h"
-#endif
 
 #include "common/log.h"
 #include "spice-client.h"
@@ -53,8 +52,6 @@
  * from the Spice client to the VM. This channel handles these messages.
  */
 
-#ifdef USE_USBREDIR
-
 #define COMPRESS_THRESHOLD 1000
 enum SpiceUsbredirChannelState {
     STATE_DISCONNECTED,
@@ -94,35 +91,25 @@ static void usbredir_log(void *user_data, int level, const char *msg);
 static int usbredir_read_callback(void *user_data, uint8_t *data, int count);
 static int usbredir_write_callback(void *user_data, uint8_t *data, int count);
 static void usbredir_write_flush_callback(void *user_data);
-#if USBREDIR_VERSION >= 0x000701
 static uint64_t usbredir_buffered_output_size_callback(void *user_data);
-#endif
 
 static void *usbredir_alloc_lock(void);
 static void usbredir_lock_lock(void *user_data);
 static void usbredir_unlock_lock(void *user_data);
 static void usbredir_free_lock(void *user_data);
 
-#else
-struct _SpiceUsbredirChannelPrivate {
-};
-#endif
-
 G_DEFINE_TYPE_WITH_PRIVATE(SpiceUsbredirChannel, spice_usbredir_channel, SPICE_TYPE_CHANNEL)
 
+
 /* ------------------------------------------------------------------ */
 
 static void spice_usbredir_channel_init(SpiceUsbredirChannel *channel)
 {
-#ifdef USE_USBREDIR
     channel->priv = spice_usbredir_channel_get_instance_private(channel);
     g_mutex_init(&channel->priv->device_connect_mutex);
-#endif
 }
 
-#ifdef USE_USBREDIR
-
-static void _channel_reset_finish(SpiceUsbredirChannel *channel)
+static void _channel_reset_finish(SpiceUsbredirChannel *channel, gboolean migrating)
 {
     SpiceUsbredirChannelPrivate *priv = channel->priv;
 
@@ -135,6 +122,8 @@ static void _channel_reset_finish(SpiceUsbredirChannel *channel)
     spice_usbredir_channel_set_context(channel, priv->context);
 
     spice_usbredir_channel_unlock(channel);
+
+    SPICE_CHANNEL_CLASS(spice_usbredir_channel_parent_class)->channel_reset(SPICE_CHANNEL(channel), migrating);
 }
 
 static void _channel_reset_cb(GObject *gobject,
@@ -146,9 +135,7 @@ static void _channel_reset_cb(GObject *gobject,
     gboolean migrating = GPOINTER_TO_UINT(user_data);
     GError *err = NULL;
 
-    _channel_reset_finish(channel);
-
-    SPICE_CHANNEL_CLASS(spice_usbredir_channel_parent_class)->channel_reset(spice_channel, migrating);
+    _channel_reset_finish(channel, migrating);
 
     spice_usbredir_channel_disconnect_device_finish(channel, result, &err);
 }
@@ -158,22 +145,30 @@ static void spice_usbredir_channel_reset(SpiceChannel *c, gboolean migrating)
     SpiceUsbredirChannel *channel = SPICE_USBREDIR_CHANNEL(c);
     SpiceUsbredirChannelPrivate *priv = channel->priv;
 
-    if (priv->host) {
-        if (priv->state == STATE_CONNECTED) {
-            spice_usbredir_channel_disconnect_device_async(channel, NULL,
-                _channel_reset_cb, GUINT_TO_POINTER(migrating));
-        } else {
-            _channel_reset_finish(channel);
-        }
-    } else {
+    /* Host isn't running, just reset */
+    if (!priv->host) {
         SPICE_CHANNEL_CLASS(spice_usbredir_channel_parent_class)->channel_reset(c, migrating);
+        return;
     }
+
+    /* Host is running, so we might need to disconnect the usb devices async.
+     * This should not block channel_reset() otherwise we might run in reconnection
+     * problems such as https://bugzilla.redhat.com/show_bug.cgi?id=1625550
+     * No operation from here on should rely on SpiceChannel as its coroutine
+     * might be terminated. */
+    
+    if (priv->state == STATE_CONNECTED) {
+        /* FIXME: We should chain-up parent's channel-reset here */
+        spice_usbredir_channel_disconnect_device_async(channel, NULL,
+            _channel_reset_cb, GUINT_TO_POINTER(migrating));
+        return;
+    }
+
+    _channel_reset_finish(channel, migrating);
 }
-#endif
 
 static void spice_usbredir_channel_class_init(SpiceUsbredirChannelClass *klass)
 {
-#ifdef USE_USBREDIR
     GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
     SpiceChannelClass *channel_class = SPICE_CHANNEL_CLASS(klass);
 
@@ -183,10 +178,8 @@ static void spice_usbredir_channel_class_init(SpiceUsbredirChannelClass *klass)
     channel_class->channel_reset = spice_usbredir_channel_reset;
 
     channel_set_handlers(SPICE_CHANNEL_CLASS(klass));
-#endif
 }
 
-#ifdef USE_USBREDIR
 static void spice_usbredir_channel_dispose(GObject *obj)
 {
     SpiceUsbredirChannel *channel = SPICE_USBREDIR_CHANNEL(obj);
@@ -229,9 +222,7 @@ static void spice_usbredir_channel_finalize(GObject *obj)
 
     if (channel->priv->host)
         usbredirhost_close(channel->priv->host);
-#ifdef USE_USBREDIR
     g_mutex_clear(&channel->priv->device_connect_mutex);
-#endif
 
     /* Chain up to the parent class */
     if (G_OBJECT_CLASS(spice_usbredir_channel_parent_class)->finalize)
@@ -276,9 +267,7 @@ void spice_usbredir_channel_set_context(SpiceUsbredirChannel *channel,
     if (!priv->host)
         g_error("Out of memory allocating usbredirhost");
 
-#if USBREDIR_VERSION >= 0x000701
     usbredirhost_set_buffered_output_size_cb(priv->host, usbredir_buffered_output_size_callback);
-#endif
 #ifdef USE_LZ4
     spice_channel_set_capability(channel, SPICE_SPICEVMC_CAP_DATA_COMPRESS_LZ4);
 #endif
@@ -352,7 +341,8 @@ static void spice_usbredir_channel_open_acl_cb(
         spice_usbredir_channel_open_device(channel, &err);
     }
     if (err) {
-        g_clear_pointer(&priv->device, libusb_unref_device);
+        libusb_unref_device(priv->device);
+        priv->device = NULL;
         g_boxed_free(spice_usb_device_get_type(), priv->spice_device);
         priv->spice_device = NULL;
         priv->state  = STATE_DISCONNECTED;
@@ -383,7 +373,8 @@ _open_device_async_cb(GTask *task,
     spice_usbredir_channel_lock(channel);
 
     if (!spice_usbredir_channel_open_device(channel, &err)) {
-        g_clear_pointer(&priv->device, libusb_unref_device);
+        libusb_unref_device(priv->device);
+        priv->device = NULL;
         g_boxed_free(spice_usb_device_get_type(), priv->spice_device);
         priv->spice_device = NULL;
     }
@@ -504,7 +495,8 @@ void spice_usbredir_channel_disconnect_device(SpiceUsbredirChannel *channel)
 
         /* This also closes the libusb handle we passed from open_device */
         usbredirhost_set_device(priv->host, NULL);
-        g_clear_pointer(&priv->device, libusb_unref_device);
+        libusb_unref_device(priv->device);
+        priv->device = NULL;
         g_boxed_free(spice_usb_device_get_type(), priv->spice_device);
         priv->spice_device = NULL;
         priv->state  = STATE_DISCONNECTED;
@@ -576,13 +568,11 @@ void spice_usbredir_channel_get_guest_filter(
 /* ------------------------------------------------------------------ */
 /* callbacks (any context)                                            */
 
-#if USBREDIR_VERSION >= 0x000701
 static uint64_t usbredir_buffered_output_size_callback(void *user_data)
 {
     g_return_val_if_fail(SPICE_IS_USBREDIR_CHANNEL(user_data), 0);
     return spice_channel_get_queue_size(SPICE_CHANNEL(user_data));
 }
-#endif
 
 /* Note that this function must be re-entrant safe, as it can get called
    from both the main thread as well as from the usb event handling thread */
@@ -940,4 +930,17 @@ static void usbredir_handle_msg(SpiceChannel *c, SpiceMsgIn *in)
     }
 }
 
+#else
+#include "spice-client.h"
+
+G_DEFINE_TYPE(SpiceUsbredirChannel, spice_usbredir_channel, SPICE_TYPE_CHANNEL)
+
+static void spice_usbredir_channel_init(SpiceUsbredirChannel *channel)
+{
+}
+
+static void spice_usbredir_channel_class_init(SpiceUsbredirChannelClass *klass)
+{
+}
+
 #endif /* USE_USBREDIR */
