@@ -482,10 +482,8 @@ spice_make_pipe(GIOStream **p1, GIOStream **p2)
     make_gio_pipe(&in1, &out2);
     make_gio_pipe(&in2, &out1);
 
-G_GNUC_BEGIN_IGNORE_DEPRECATIONS
     *p1 = g_simple_io_stream_new(in1, out1);
     *p2 = g_simple_io_stream_new(in2, out2);
-G_GNUC_END_IGNORE_DEPRECATIONS
 
     g_object_unref(in1);
     g_object_unref(in2);
