@@ -33,7 +33,7 @@
 #include "gio-coroutine.h"
 
 #include "common/client_marshallers.h"
-#include "common/client_demarshallers.h"
+#include "common/demarshallers.h"
 #include "common/ssl_verify.h"
 
 G_BEGIN_DECLS
@@ -129,7 +129,6 @@ struct _SpiceChannelPrivate {
     SpiceLinkHeader             link_hdr;
     SpiceLinkHeader             peer_hdr;
     SpiceLinkReply*             peer_msg;
-    int                         peer_pos;
 
     int                         message_ack_window;
     int                         message_ack_count;
