@@ -51,11 +51,11 @@ static void spice_channel_handle_msg(SpiceChannel *channel, SpiceMsgIn *msg);
 static void spice_channel_write_msg(SpiceChannel *channel, SpiceMsgOut *out);
 static void spice_channel_send_link(SpiceChannel *channel);
 static void channel_reset(SpiceChannel *channel, gboolean migrating);
-static void spice_channel_reset_capabilities(SpiceChannel *channel);
 static void spice_channel_send_migration_handshake(SpiceChannel *channel);
 static gboolean channel_connect(SpiceChannel *channel, gboolean tls);
 
-#if OPENSSL_VERSION_NUMBER < 0x10100000 || defined(LIBRESSL_VERSION_NUMBER)
+#if OPENSSL_VERSION_NUMBER < 0x10100000 || \
+    (defined(LIBRESSL_VERSION_NUMBER) && LIBRESSL_VERSION_NUMBER < 0x20700000)
 static RSA *EVP_PKEY_get0_RSA(EVP_PKEY *pkey)
 {
     if (pkey->type != EVP_PKEY_RSA) {
@@ -80,7 +80,7 @@ static RSA *EVP_PKEY_get0_RSA(EVP_PKEY *pkey)
 
 G_DEFINE_TYPE_WITH_CODE (SpiceChannel, spice_channel, G_TYPE_OBJECT,
                          G_ADD_PRIVATE (SpiceChannel)
-                         g_type_add_class_private (g_define_type_id, sizeof (SpiceChannelClassPrivate)));
+                         g_type_add_class_private (g_define_type_id, sizeof (SpiceChannelClassPrivate)))
 
 /* Properties */
 enum {
@@ -1059,7 +1059,7 @@ static int spice_channel_read_wire(SpiceChannel *channel, void *data, size_t len
                 continue;
             } else {
                 c->has_error = TRUE;
-                return -errno;
+                return errno > 0 ? -errno : -EIO;
             }
         }
         if (ret == 0) {
@@ -1123,6 +1123,8 @@ static int spice_channel_read_sasl(SpiceChannel *channel, void *data, size_t len
 
 /*
  * Fill the 'data' buffer up with exactly 'len' bytes worth of data
+ * Returns 0 if connection was closed or on unknown errors, <0 for error and
+ * length on success
  */
 /* coroutine context */
 static int spice_channel_read(SpiceChannel *channel, void *data, size_t length)
@@ -1314,11 +1316,8 @@ static void spice_channel_send_link(SpiceChannel *channel)
     g_object_get(c->session, "protocol", &protocol, NULL);
     switch (protocol) {
     case 1: /* protocol 1 == major 1, old 0.4 protocol, last active minor */
-        c->link_hdr.major_version = 1;
-        c->link_hdr.minor_version = 3;
-        c->parser = spice_get_server_channel_parser1(c->channel_type, NULL);
-        c->marshallers = spice_message_marshallers_get1();
-        break;
+        g_critical("deprecated major %d", protocol);
+        return;
     case SPICE_VERSION_MAJOR: /* protocol 2 == current */
         c->link_hdr.major_version = SPICE_VERSION_MAJOR;
         c->link_hdr.minor_version = SPICE_VERSION_MINOR;
@@ -1400,6 +1399,11 @@ static gboolean spice_channel_recv_link_hdr(SpiceChannel *channel)
     c->peer_hdr.minor_version = GUINT32_FROM_LE(c->peer_hdr.minor_version);
     c->peer_hdr.size = GUINT32_FROM_LE(c->peer_hdr.size);
 
+    if (c->peer_hdr.size < sizeof(*c->peer_msg)) {
+        g_warning("invalid peer header size: %u", c->peer_hdr.size);
+        goto error;
+    }
+
     c->peer_msg = g_malloc0(c->peer_hdr.size);
     if (c->peer_msg == NULL) {
         g_warning("invalid peer header size: %u", c->peer_hdr.size);
@@ -1409,16 +1413,6 @@ static gboolean spice_channel_recv_link_hdr(SpiceChannel *channel)
     return TRUE;
 
 error:
-    /* Windows socket seems to give early CONNRESET errors. The server
-       does not linger when closing the socket if the protocol is
-       incompatible. Try with the oldest protocol in this case: */
-    if (c->link_hdr.major_version != 1) {
-        SPICE_DEBUG("%s: error, switching to protocol 1 (spice 0.4)", c->name);
-        c->state = SPICE_CHANNEL_STATE_RECONNECTING;
-        g_object_set(c->session, "protocol", 1, NULL);
-        return FALSE;
-    }
-
     c->event = SPICE_CHANNEL_ERROR_LINK;
     return FALSE;
 }
@@ -1908,9 +1902,10 @@ static void store_caps(const uint8_t *caps_src, uint32_t ncaps,
 static gboolean spice_channel_recv_link_msg(SpiceChannel *channel)
 {
     SpiceChannelPrivate *c;
-    int rc, num_caps;
+    int rc;
+    uint32_t num_caps;
     uint32_t num_channel_caps, num_common_caps;
-    uint8_t *caps_src;
+    const uint8_t *caps_src, *caps_end;
     SpiceChannelEvent event = SPICE_CHANNEL_ERROR_LINK;
 
     g_return_val_if_fail(channel != NULL, FALSE);
@@ -1918,14 +1913,18 @@ static gboolean spice_channel_recv_link_msg(SpiceChannel *channel)
 
     c = channel->priv;
 
-    rc = spice_channel_read(channel, (uint8_t*)c->peer_msg + c->peer_pos,
-                            c->peer_hdr.size - c->peer_pos);
-    c->peer_pos += rc;
-    if (c->peer_pos != c->peer_hdr.size) {
+    rc = spice_channel_read(channel, (uint8_t*)c->peer_msg, c->peer_hdr.size);
+    if (rc != c->peer_hdr.size) {
         g_critical("%s: %s: incomplete link reply (%d/%u)",
                   c->name, __FUNCTION__, rc, c->peer_hdr.size);
         goto error;
     }
+
+    c->peer_msg->error = GUINT32_FROM_LE(c->peer_msg->error);
+    c->peer_msg->num_common_caps = GUINT32_FROM_LE(c->peer_msg->num_common_caps);
+    c->peer_msg->num_channel_caps = GUINT32_FROM_LE(c->peer_msg->num_channel_caps);
+    c->peer_msg->caps_offset = GUINT32_FROM_LE(c->peer_msg->caps_offset);
+
     switch (c->peer_msg->error) {
     case SPICE_LINK_ERR_OK:
         /* nothing */
@@ -1941,20 +1940,31 @@ static gboolean spice_channel_recv_link_msg(SpiceChannel *channel)
         goto error;
     }
 
-    num_channel_caps = GUINT32_FROM_LE(c->peer_msg->num_channel_caps);
-    num_common_caps = GUINT32_FROM_LE(c->peer_msg->num_common_caps);
+    num_channel_caps = c->peer_msg->num_channel_caps;
+    num_common_caps = c->peer_msg->num_common_caps;
 
     num_caps = num_channel_caps + num_common_caps;
-    CHANNEL_DEBUG(channel, "%s: %d caps", __FUNCTION__, num_caps);
+    CHANNEL_DEBUG(channel, "%s: %u caps", __FUNCTION__, num_caps);
+
+    if (c->peer_msg->caps_offset > c->peer_hdr.size) {
+        goto error;
+    }
+    caps_end = (uint8_t*)c->peer_msg + c->peer_hdr.size;
 
     /* see original spice/client code: */
     /* g_return_if_fail(c->peer_msg + c->peer_msg->caps_offset * sizeof(uint32_t) > c->peer_msg + c->peer_hdr.size); */
 
-    caps_src = (uint8_t *)c->peer_msg + GUINT32_FROM_LE(c->peer_msg->caps_offset);
+    caps_src = (uint8_t *)c->peer_msg + c->peer_msg->caps_offset;
+    if ((caps_end - caps_src) / sizeof(uint32_t) < num_common_caps) {
+        goto error;
+    }
     CHANNEL_DEBUG(channel, "got remote common caps:");
     store_caps(caps_src, num_common_caps, c->remote_common_caps);
 
     caps_src += num_common_caps * sizeof(uint32_t);
+    if ((caps_end - caps_src) / sizeof(uint32_t) < num_channel_caps) {
+        goto error;
+    }
     CHANNEL_DEBUG(channel, "got remote channel caps:");
     store_caps(caps_src, num_channel_caps, c->remote_caps);
 
@@ -2674,11 +2684,15 @@ cleanup:
     if (c->state == SPICE_CHANNEL_STATE_RECONNECTING ||
         c->state == SPICE_CHANNEL_STATE_SWITCHING) {
         g_warn_if_fail(c->event == SPICE_CHANNEL_NONE);
-        channel_connect(channel, c->tls);
-        g_object_unref(channel);
-    } else
-        g_idle_add(spice_channel_delayed_unref, data);
+        if (channel_connect(channel, c->tls)) {
+            g_object_unref(channel);
+            return NULL;
+        }
 
+        c->event = SPICE_CHANNEL_ERROR_CONNECT;
+    }
+
+    g_idle_add(spice_channel_delayed_unref, channel);
     /* Co-routine exits now - the SpiceChannel object may no longer exist,
        so don't do anything else now unless you like SEGVs */
     return NULL;
@@ -2829,7 +2843,6 @@ static void channel_reset(SpiceChannel *channel, gboolean migrating)
     c->auth_needs_password = FALSE;
 
     g_clear_pointer(&c->peer_msg, g_free);
-    c->peer_pos = 0;
 
     g_mutex_lock(&c->xmit_queue_lock);
     c->xmit_queue_blocked = TRUE; /* Disallow queuing new messages */
@@ -2845,11 +2858,6 @@ static void channel_reset(SpiceChannel *channel, gboolean migrating)
 
     g_array_set_size(c->remote_common_caps, 0);
     g_array_set_size(c->remote_caps, 0);
-    g_array_set_size(c->common_caps, 0);
-    /* Restore our default capabilities in case the channel gets re-used */
-    spice_channel_set_common_capability(channel, SPICE_COMMON_CAP_PROTOCOL_AUTH_SELECTION);
-    spice_channel_set_common_capability(channel, SPICE_COMMON_CAP_MINI_HEADER);
-    spice_channel_reset_capabilities(channel);
 
     if (c->state == SPICE_CHANNEL_STATE_SWITCHING)
         spice_session_set_migration_state(spice_channel_get_session(channel),
@@ -3092,16 +3100,6 @@ static void spice_channel_handle_msg(SpiceChannel *channel, SpiceMsgIn *msg)
     handler(channel, msg);
 }
 
-static void spice_channel_reset_capabilities(SpiceChannel *channel)
-{
-    SpiceChannelPrivate *c = channel->priv;
-    g_array_set_size(c->caps, 0);
-
-    if (SPICE_CHANNEL_GET_CLASS(channel)->channel_reset_capabilities) {
-        SPICE_CHANNEL_GET_CLASS(channel)->channel_reset_capabilities(channel);
-    }
-}
-
 static void spice_channel_send_migration_handshake(SpiceChannel *channel)
 {
     SpiceChannelPrivate *c = channel->priv;
