@@ -59,11 +59,6 @@ static void spice_gstaudio_get_record_volume_info_async(SpiceAudio *audio,
 static gboolean spice_gstaudio_get_record_volume_info_finish(SpiceAudio *audio,
         GAsyncResult *res, gboolean *mute, guint8 *nchannels, guint16 **volume, GError **error);
 
-static void spice_gstaudio_finalize(GObject *obj)
-{
-    G_OBJECT_CLASS(spice_gstaudio_parent_class)->finalize(obj);
-}
-
 static void stream_dispose(struct stream *s)
 {
     if (s->pipe) {
@@ -113,7 +108,6 @@ static void spice_gstaudio_class_init(SpiceGstaudioClass *klass)
     audio_class->get_record_volume_info_async = spice_gstaudio_get_record_volume_info_async;
     audio_class->get_record_volume_info_finish = spice_gstaudio_get_record_volume_info_finish;
 
-    gobject_class->finalize = spice_gstaudio_finalize;
     gobject_class->dispose = spice_gstaudio_dispose;
 }
 
@@ -197,7 +191,7 @@ static void record_start(SpiceRecordChannel *channel, gint format, gint channels
     if (p->record.pipe &&
         (p->record.rate != frequency ||
          p->record.channels != channels)) {
-        record_stop(gstaudio);
+        gst_element_set_state(p->record.pipe, GST_STATE_NULL);
         g_clear_pointer(&p->record.pipe, gst_object_unref);
     }
 
@@ -350,7 +344,7 @@ static void playback_data(SpicePlaybackChannel *channel,
 static void playback_volume_changed(GObject *object, GParamSpec *pspec, gpointer data)
 {
     SpiceGstaudio *gstaudio = data;
-    GstElement *e;
+    GstElement *e = NULL;
     guint16 *volume;
     guint nchannels;
     SpiceGstaudioPrivate *p = gstaudio->priv;
@@ -371,13 +365,19 @@ static void playback_volume_changed(GObject *object, GParamSpec *pspec, gpointer
 
     if (GST_IS_BIN(p->playback.sink))
         e = gst_bin_get_by_interface(GST_BIN(p->playback.sink), GST_TYPE_STREAM_VOLUME);
-    else
+    if (!e)
         e = g_object_ref(p->playback.sink);
 
-    if (GST_IS_STREAM_VOLUME(e))
+    /* Double check to make compiler happy */
+    g_return_if_fail(e != NULL);
+
+    if (GST_IS_STREAM_VOLUME(e)) {
         gst_stream_volume_set_volume(GST_STREAM_VOLUME(e), GST_STREAM_VOLUME_FORMAT_CUBIC, vol);
-    else
+    } else if (g_object_class_find_property(G_OBJECT_GET_CLASS (e), "volume") != NULL) {
         g_object_set(e, "volume", vol, NULL);
+    } else {
+        g_warning("playback: ignoring volume change on %s", gst_element_get_name(e));
+    }
 
     g_object_unref(e);
 }
@@ -386,7 +386,7 @@ static void playback_mute_changed(GObject *object, GParamSpec *pspec, gpointer d
 {
     SpiceGstaudio *gstaudio = data;
     SpiceGstaudioPrivate *p = gstaudio->priv;
-    GstElement *e;
+    GstElement *e = NULL;
     gboolean mute;
 
     if (!p->playback.sink)
@@ -397,11 +397,19 @@ static void playback_mute_changed(GObject *object, GParamSpec *pspec, gpointer d
 
     if (GST_IS_BIN(p->playback.sink))
         e = gst_bin_get_by_interface(GST_BIN(p->playback.sink), GST_TYPE_STREAM_VOLUME);
-    else
+    if (!e)
         e = g_object_ref(p->playback.sink);
 
-    if (GST_IS_STREAM_VOLUME(e))
+    /* Double check to make compiler happy */
+    g_return_if_fail(e != NULL);
+
+    if (GST_IS_STREAM_VOLUME(e)) {
         gst_stream_volume_set_mute(GST_STREAM_VOLUME(e), mute);
+    } else if (g_object_class_find_property(G_OBJECT_GET_CLASS (e), "mute") != NULL) {
+        g_object_set(e, "mute", mute, NULL);
+    } else {
+        g_warning("playback: ignoring mute change on %s", gst_element_get_name(e));
+    }
 
     g_object_unref(e);
 }
@@ -410,7 +418,7 @@ static void record_volume_changed(GObject *object, GParamSpec *pspec, gpointer d
 {
     SpiceGstaudio *gstaudio = data;
     SpiceGstaudioPrivate *p = gstaudio->priv;
-    GstElement *e;
+    GstElement *e = NULL;
     guint16 *volume;
     guint nchannels;
     gdouble vol;
@@ -430,13 +438,19 @@ static void record_volume_changed(GObject *object, GParamSpec *pspec, gpointer d
 
     if (GST_IS_BIN(p->record.src))
         e = gst_bin_get_by_interface(GST_BIN(p->record.src), GST_TYPE_STREAM_VOLUME);
-    else
+    if (!e)
         e = g_object_ref(p->record.src);
 
-    if (GST_IS_STREAM_VOLUME(e))
+    /* Double check to make compiler happy */
+    g_return_if_fail(e != NULL);
+
+    if (GST_IS_STREAM_VOLUME(e)) {
         gst_stream_volume_set_volume(GST_STREAM_VOLUME(e), GST_STREAM_VOLUME_FORMAT_CUBIC, vol);
-    else
-        g_warning("gst lacks volume capabilities on src");
+    } else if (g_object_class_find_property(G_OBJECT_GET_CLASS (e), "volume") != NULL) {
+        g_object_set(e, "volume", vol, NULL);
+    } else {
+        g_warning("record: ignoring volume change on %s", gst_element_get_name(e));
+    }
 
     g_object_unref(e);
 }
@@ -445,7 +459,7 @@ static void record_mute_changed(GObject *object, GParamSpec *pspec, gpointer dat
 {
     SpiceGstaudio *gstaudio = data;
     SpiceGstaudioPrivate *p = gstaudio->priv;
-    GstElement *e;
+    GstElement *e = NULL;
     gboolean mute;
 
     if (!p->record.src)
@@ -456,13 +470,19 @@ static void record_mute_changed(GObject *object, GParamSpec *pspec, gpointer dat
 
     if (GST_IS_BIN(p->record.src))
         e = gst_bin_get_by_interface(GST_BIN(p->record.src), GST_TYPE_STREAM_VOLUME);
-    else
+    if (!e)
         e = g_object_ref(p->record.src);
 
-    if (GST_IS_STREAM_VOLUME (e))
+    /* Double check to make compiler happy */
+    g_return_if_fail(e != NULL);
+
+    if (GST_IS_STREAM_VOLUME (e)) {
         gst_stream_volume_set_mute(GST_STREAM_VOLUME(e), mute);
-    else
-        g_warning("gst lacks mute capabilities on src: %d", mute);
+    } else if (g_object_class_find_property(G_OBJECT_GET_CLASS (e), "mute") != NULL) {
+        g_object_set(e, "mute", mute, NULL);
+    } else {
+        g_warning("record: ignoring mute change on %s", gst_element_get_name(e));
+    }
 
     g_object_unref(e);
 }
@@ -533,7 +553,35 @@ SpiceGstaudio *spice_gstaudio_new(SpiceSession *session, GMainContext *context,
                                   const char *name)
 {
     GError *err = NULL;
+
     if (gst_init_check(NULL, NULL, &err)) {
+        GstPluginFeature *pulsesrc;
+
+        pulsesrc = gst_registry_lookup_feature(gst_registry_get(), "pulsesrc");
+        if (pulsesrc) {
+            unsigned major, minor, micro;
+            GstPlugin *plugin = gst_plugin_feature_get_plugin(pulsesrc);
+
+            if (sscanf(gst_plugin_get_version(plugin), "%u.%u.%u",
+                       &major, &minor, &micro) != 3) {
+                g_warn_if_reached();
+                gst_object_unref(plugin);
+                gst_object_unref(pulsesrc);
+                return NULL;
+            }
+
+            if (major < 1 ||
+                (major == 1 && minor < 14) ||
+                (major == 1 && minor == 14 && micro < 5)) {
+                g_warning("Bad pulsesrc version %s, lowering its rank",
+                          gst_plugin_get_version(plugin));
+                gst_plugin_feature_set_rank(pulsesrc, GST_RANK_NONE);
+            }
+
+            gst_object_unref(plugin);
+            gst_object_unref(pulsesrc);
+        }
+
         return g_object_new(SPICE_TYPE_GSTAUDIO,
                             "session", session,
                             "main-context", context,
@@ -564,7 +612,7 @@ static gboolean spice_gstaudio_get_playback_volume_info_finish(SpiceAudio *audio
                                                                GError **error)
 {
     SpiceGstaudioPrivate *p = SPICE_GSTAUDIO(audio)->priv;
-    GstElement *e;
+    GstElement *e = NULL;
     gboolean lmute;
     gdouble vol;
     GTask *task = G_TASK(res);
@@ -588,7 +636,7 @@ static gboolean spice_gstaudio_get_playback_volume_info_finish(SpiceAudio *audio
 
     if (GST_IS_BIN(p->playback.sink))
         e = gst_bin_get_by_interface(GST_BIN(p->playback.sink), GST_TYPE_STREAM_VOLUME);
-    else
+    if (!e)
         e = g_object_ref(p->playback.sink);
 
     if (GST_IS_STREAM_VOLUME(e)) {
@@ -646,7 +694,7 @@ static gboolean spice_gstaudio_get_record_volume_info_finish(SpiceAudio *audio,
                                                              GError **error)
 {
     SpiceGstaudioPrivate *p = SPICE_GSTAUDIO(audio)->priv;
-    GstElement *e;
+    GstElement *e = NULL;
     gboolean lmute;
     gdouble vol;
     gboolean fake_channel = FALSE;
@@ -671,7 +719,7 @@ static gboolean spice_gstaudio_get_record_volume_info_finish(SpiceAudio *audio,
 
     if (GST_IS_BIN(p->record.src))
         e = gst_bin_get_by_interface(GST_BIN(p->record.src), GST_TYPE_STREAM_VOLUME);
-    else
+    if (!e)
         e = g_object_ref(p->record.src);
 
     if (GST_IS_STREAM_VOLUME(e)) {
