@@ -104,7 +104,7 @@ static void channel_destroy(SpiceSession *session, SpiceChannel *channel,
                             gpointer user_data);
 static gboolean read_only(SpiceGtkSession *self);
 
-G_DEFINE_TYPE_WITH_PRIVATE(SpiceGtkSession, spice_gtk_session, G_TYPE_OBJECT);
+G_DEFINE_TYPE_WITH_PRIVATE(SpiceGtkSession, spice_gtk_session, G_TYPE_OBJECT)
 
 /* Properties */
 enum {
@@ -119,10 +119,8 @@ enum {
 static guint32 get_keyboard_lock_modifiers(void)
 {
     guint32 modifiers = 0;
-#if GTK_CHECK_VERSION(3,18,0)
 /* Ignore GLib's too-new warnings */
-G_GNUC_BEGIN_IGNORE_DEPRECATIONS
-    GdkKeymap *keyboard = gdk_keymap_get_default();
+    GdkKeymap *keyboard = gdk_keymap_get_for_display(gdk_display_get_default());
 
     if (gdk_keymap_get_caps_lock_state(keyboard)) {
         modifiers |= SPICE_INPUTS_CAPS_LOCK;
@@ -135,44 +133,6 @@ G_GNUC_BEGIN_IGNORE_DEPRECATIONS
     if (gdk_keymap_get_scroll_lock_state(keyboard)) {
         modifiers |= SPICE_INPUTS_SCROLL_LOCK;
     }
-G_GNUC_END_IGNORE_DEPRECATIONS
-#else
-#ifdef HAVE_X11_XKBLIB_H
-    Display *x_display = NULL;
-    XKeyboardState keyboard_state;
-
-    GdkScreen *screen = gdk_screen_get_default();
-    if (!GDK_IS_X11_DISPLAY(gdk_screen_get_display(screen))) {
-        SPICE_DEBUG("FIXME: gtk backend is not X11");
-        return 0;
-    }
-
-    x_display = GDK_SCREEN_XDISPLAY(screen);
-    XGetKeyboardControl(x_display, &keyboard_state);
-
-    if (keyboard_state.led_mask & 0x01) {
-        modifiers |= SPICE_INPUTS_CAPS_LOCK;
-    }
-    if (keyboard_state.led_mask & 0x02) {
-        modifiers |= SPICE_INPUTS_NUM_LOCK;
-    }
-    if (keyboard_state.led_mask & 0x04) {
-        modifiers |= SPICE_INPUTS_SCROLL_LOCK;
-    }
-#elif defined(G_OS_WIN32)
-    if (GetKeyState(VK_CAPITAL) & 1) {
-        modifiers |= SPICE_INPUTS_CAPS_LOCK;
-    }
-    if (GetKeyState(VK_NUMLOCK) & 1) {
-        modifiers |= SPICE_INPUTS_NUM_LOCK;
-    }
-    if (GetKeyState(VK_SCROLL) & 1) {
-        modifiers |= SPICE_INPUTS_SCROLL_LOCK;
-    }
-#else
-    g_warning("get_keyboard_lock_modifiers not implemented");
-#endif // HAVE_X11_XKBLIB_H
-#endif // GTK_CHECK_VERSION(3,18,0)
     return modifiers;
 }
 
@@ -203,6 +163,14 @@ static void keymap_modifiers_changed(GdkKeymap *keymap, gpointer data)
 {
     SpiceGtkSession *self = data;
 
+    /* set_key_locks() is inherently racy, but no need to sync modifiers
+     * if we have focus as the regular keypress/keyrelease will have set
+     * the expected modifiers state in the guest.
+     */
+    if (self->priv->keyboard_has_focus) {
+        return;
+    }
+
     spice_gtk_session_sync_keyboard_modifiers(self);
 }
 
@@ -216,7 +184,7 @@ static void guest_modifiers_changed(SpiceInputsChannel *inputs, gpointer data)
 static void spice_gtk_session_init(SpiceGtkSession *self)
 {
     SpiceGtkSessionPrivate *s;
-    GdkKeymap *keymap = gdk_keymap_get_default();
+    GdkKeymap *keymap = gdk_keymap_get_for_display(gdk_display_get_default());
 
     s = self->priv = spice_gtk_session_get_instance_private(self);
 
@@ -658,6 +626,23 @@ static void clipboard_get_targets(GtkClipboard *clipboard,
     s->nclip_targets[selection] = 0;
 }
 
+/* Callback for every owner-change event for given @clipboard.
+ * This event is triggered in different ways depending on the environment of
+ * the Client, some examples:
+ *
+ * Situation 1: When another application on the client machine is holding and
+ * changing the clipboard. If client is on Wayland, spice-gtk only receives the
+ * related GtkClipboard::owner-changed event after focus-in event on Spice
+ * widget; On X11, we will receive it at the moment the clipboard data has been
+ * changed in by other application.
+ *
+ * Situation 2: When spice-gtk holds the focus and is changing the clipboard by
+ * either setting new content information with gtk_clipboard_set_with_owner() or
+ * clearing up old content with gtk_clipboard_clear(). The main difference between
+ * Wayland and X11 is that on X11, gtk_clipboard_clear() sets the owner to none, which
+ * emits owner-change event; On Wayland that does not happen as spice-gtk still is
+ * the owner of the clipboard.
+ */
 static void clipboard_owner_change(GtkClipboard        *clipboard,
                                    GdkEventOwnerChange *event,
                                    gpointer            user_data)
@@ -671,30 +656,44 @@ static void clipboard_owner_change(GtkClipboard        *clipboard,
     selection = get_selection_from_clipboard(s, clipboard);
     g_return_if_fail(selection != -1);
 
-    if (s->main == NULL)
+    if (s->main == NULL) {
         return;
+    }
 
+    /* In case we sent a grab to the agent, we need to release it now as
+     * previous clipboard data should not be reachable anymore */
     if (s->clip_grabbed[selection]) {
         s->clip_grabbed[selection] = FALSE;
-        if (spice_main_channel_agent_test_capability(s->main, VD_AGENT_CAP_CLIPBOARD_BY_DEMAND))
+        if (spice_main_channel_agent_test_capability(s->main, VD_AGENT_CAP_CLIPBOARD_BY_DEMAND)) {
             spice_main_channel_clipboard_selection_release(s->main, selection);
+        }
     }
 
-    switch (event->reason) {
-    case GDK_OWNER_CHANGE_NEW_OWNER:
-        if (gtk_clipboard_get_owner(clipboard) == G_OBJECT(self))
-            break;
-
-        s->clipboard_by_guest[selection] = FALSE;
-        s->clip_hasdata[selection] = TRUE;
-        if (s->auto_clipboard_enable && !read_only(self))
-            gtk_clipboard_request_targets(clipboard, clipboard_get_targets,
-                                          get_weak_ref(self));
-        break;
-    default:
+    /* We are mostly interested when owner has changed in which case
+     * we would like to let agent know about new clipboard data. */
+    if (event->reason != GDK_OWNER_CHANGE_NEW_OWNER) {
         s->clip_hasdata[selection] = FALSE;
-        break;
+        return;
     }
+
+    /* This situation happens when clipboard is being set by us (grab message) */
+    if (gtk_clipboard_get_owner(clipboard) == G_OBJECT(self)) {
+        return;
+    }
+
+    s->clipboard_by_guest[selection] = FALSE;
+
+#ifdef GDK_WINDOWING_X11
+    if (!event->owner && GDK_IS_X11_DISPLAY(gdk_display_get_default())) {
+        s->clip_hasdata[selection] = FALSE;
+        return;
+    }
+#endif
+
+    s->clip_hasdata[selection] = TRUE;
+    if (s->auto_clipboard_enable && !read_only(self))
+        gtk_clipboard_request_targets(clipboard, clipboard_get_targets,
+                                      get_weak_ref(self));
 }
 
 typedef struct
@@ -826,22 +825,22 @@ static gboolean clipboard_grab(SpiceMainChannel *main, guint selection,
     gboolean target_selected[SPICE_N_ELEMENTS(atom2agent)] = { FALSE, };
     gboolean found;
     GtkClipboard* cb;
-    int m, n, i;
+    int m, n;
+    int num_targets = 0;
 
     cb = get_clipboard_from_selection(s, selection);
     g_return_val_if_fail(cb != NULL, FALSE);
 
-    i = 0;
     for (n = 0; n < ntypes; ++n) {
         found = FALSE;
         for (m = 0; m < SPICE_N_ELEMENTS(atom2agent); m++) {
             if (atom2agent[m].vdagent == types[n] && !target_selected[m]) {
                 found = TRUE;
-                g_return_val_if_fail(i < SPICE_N_ELEMENTS(atom2agent), FALSE);
-                targets[i].target = (gchar*)atom2agent[m].xatom;
-                targets[i].info = m;
+                g_return_val_if_fail(num_targets < SPICE_N_ELEMENTS(atom2agent), FALSE);
+                targets[num_targets].target = (gchar*)atom2agent[m].xatom;
+                targets[num_targets].info = m;
                 target_selected[m] = TRUE;
-                i += 1;
+                num_targets++;
             }
         }
         if (!found) {
@@ -851,25 +850,29 @@ static gboolean clipboard_grab(SpiceMainChannel *main, guint selection,
     }
 
     g_free(s->clip_targets[selection]);
-    s->nclip_targets[selection] = i;
-    s->clip_targets[selection] = g_memdup(targets, sizeof(GtkTargetEntry) * i);
+    s->nclip_targets[selection] = num_targets;
+    s->clip_targets[selection] = g_memdup(targets, sizeof(GtkTargetEntry) * num_targets);
     /* Receiving a grab implies we've released our own grab */
     s->clip_grabbed[selection] = FALSE;
 
     if (read_only(self) ||
         !s->auto_clipboard_enable ||
-        s->nclip_targets[selection] == 0)
-        goto skip_grab_clipboard;
+        s->nclip_targets[selection] == 0) {
+        return TRUE;
+    }
 
-    if (!gtk_clipboard_set_with_owner(cb, targets, i,
-                                      clipboard_get, clipboard_clear, G_OBJECT(self))) {
+    if (!gtk_clipboard_set_with_owner(cb,
+                                      targets,
+                                      num_targets,
+                                      clipboard_get,
+                                      clipboard_clear,
+                                      G_OBJECT(self))) {
         g_warning("clipboard grab failed");
         return FALSE;
     }
     s->clipboard_by_guest[selection] = TRUE;
     s->clip_hasdata[selection] = FALSE;
 
-skip_grab_clipboard:
     return TRUE;
 }
 
@@ -1027,15 +1030,14 @@ static gboolean clipboard_request(SpiceMainChannel *main, guint selection,
     GtkClipboard* cb;
     int m;
 
+    cb = get_clipboard_from_selection(s, selection);
+    g_return_val_if_fail(cb != NULL, FALSE);
     g_return_val_if_fail(s->clipboard_by_guest[selection] == FALSE, FALSE);
     g_return_val_if_fail(s->clip_grabbed[selection], FALSE);
 
     if (read_only(self))
         return FALSE;
 
-    cb = get_clipboard_from_selection(s, selection);
-    g_return_val_if_fail(cb != NULL, FALSE);
-
     if (type == VD_AGENT_CLIPBOARD_UTF8_TEXT) {
         gtk_clipboard_request_text(cb, clipboard_received_text_cb,
                                    get_weak_ref(self));
