@@ -17,12 +17,13 @@
 */
 #include "config.h"
 
+/* include first, on Windows will override winsock definitions */
+#include <gio/gnetworking.h>
 #include <gio/gio.h>
 #include <glib.h>
 #ifdef G_OS_UNIX
 #include <gio/gunixsocketaddress.h>
 #endif
-#include "common/ring.h"
 
 #include "spice-client.h"
 #include "spice-common.h"
@@ -34,10 +35,12 @@
 #include "channel-playback-priv.h"
 #include "spice-audio-priv.h"
 
-struct channel {
-    SpiceChannel      *channel;
-    RingItem          link;
-};
+#if !defined(SOL_TCP) && defined(IPPROTO_TCP)
+#define SOL_TCP IPPROTO_TCP
+#endif
+#if !defined(TCP_KEEPIDLE) && defined(TCP_KEEPALIVE) && defined(__APPLE__)
+#define TCP_KEEPIDLE TCP_KEEPALIVE
+#endif
 
 #define IMAGES_CACHE_SIZE_DEFAULT (1024 * 1024 * 80)
 #define MIN_GLZ_WINDOW_SIZE_DEFAULT (1024 * 1024 * 12)
@@ -67,6 +70,9 @@ struct _SpiceSessionPrivate {
     /* whether to enable smartcard event forwarding to the server */
     gboolean          smartcard;
 
+    /* whether to enable GL scanout */
+    gboolean          gl_scanout;
+
     /* list of certificates to use for the software smartcard reader if
      * enabled. For now, it has to contain exactly 3 certificates for
      * the software reader to be functional
@@ -88,16 +94,14 @@ struct _SpiceSessionPrivate {
 
     GStrv             disable_effects;
     GStrv             secure_channels;
-    gint              color_depth;
 
     int               connection_id;
     int               protocol;
     SpiceChannel      *cmain; /* weak reference */
-    Ring              channels;
+    GList             *channels;
     guint             channels_destroying;
-    guint32           mm_time;
     gboolean          client_provided_sockets;
-    guint64           mm_time_at_clock;
+    guint64           mm_time_offset;
     SpiceSession      *migration;
     GList             *migration_left;
     SpiceSessionMigration migration_state;
@@ -157,7 +161,7 @@ struct _SpiceSessionPrivate {
  * a Spice server.
  */
 
-G_DEFINE_TYPE_WITH_PRIVATE(SpiceSession, spice_session, G_TYPE_OBJECT);
+G_DEFINE_TYPE_WITH_PRIVATE(SpiceSession, spice_session, G_TYPE_OBJECT)
 
 /* Properties */
 enum {
@@ -198,6 +202,7 @@ enum {
     PROP_USERNAME,
     PROP_UNIX_PATH,
     PROP_PREF_COMPRESSION,
+    PROP_GL_SCANOUT,
 };
 
 /* signals */
@@ -227,6 +232,9 @@ static const GEnumValue _spice_image_compress_values[] = {
 
 G_STATIC_ASSERT(G_N_ELEMENTS(_spice_image_compress_values) == SPICE_IMAGE_COMPRESSION_ENUM_END + 1);
 
+static const gchar* spice_session_get_shared_dir(SpiceSession *session);
+static void spice_session_set_shared_dir(SpiceSession *session, const gchar *dir);
+
 GType
 spice_image_compress_get_type (void)
 {
@@ -284,7 +292,6 @@ static void spice_session_init(SpiceSession *session)
     SPICE_DEBUG("Supported channels: %s", channels);
     g_free(channels);
 
-    ring_init(&s->channels);
     s->images = cache_image_new((GDestroyNotify)pixman_image_unref);
     s->glz_window = glz_decoder_window_new();
     update_proxy(session, NULL);
@@ -294,19 +301,17 @@ static void
 session_disconnect(SpiceSession *self, gboolean keep_main)
 {
     SpiceSessionPrivate *s;
-    struct channel *item;
-    RingItem *ring, *next;
 
     s = self->priv;
 
-    for (ring = ring_get_head(&s->channels); ring != NULL; ring = next) {
-        next = ring_next(&s->channels, ring);
-        item = SPICE_CONTAINEROF(ring, struct channel, link);
+    for (GList *l = s->channels; l != NULL; ) {
+        SpiceChannel *channel = l->data;
+        l = l->next;
 
-        if (keep_main && item->channel == s->cmain) {
-            spice_channel_disconnect(item->channel, SPICE_CHANNEL_NONE);
+        if (keep_main && channel == s->cmain) {
+            spice_channel_disconnect(channel, SPICE_CHANNEL_NONE);
         } else {
-            spice_session_channel_destroy(self, item->channel);
+            spice_session_channel_destroy(self, channel);
         }
     }
 
@@ -333,7 +338,7 @@ spice_session_dispose(GObject *gobject)
     g_warn_if_fail(s->after_main_init == 0);
     g_warn_if_fail(s->disconnecting == 0);
     g_warn_if_fail(s->channels_destroying == 0);
-    g_warn_if_fail(ring_is_empty(&s->channels));
+    g_warn_if_fail(s->channels == NULL);
 
     g_clear_object(&s->audio_manager);
     g_clear_object(&s->usb_manager);
@@ -673,8 +678,8 @@ static void spice_session_get_property(GObject    *gobject,
     case PROP_SECURE_CHANNELS:
         g_value_set_boxed(value, s->secure_channels);
         break;
-    case PROP_COLOR_DEPTH:
-        g_value_set_int(value, s->color_depth);
+    case PROP_COLOR_DEPTH: /* FIXME: deprecated */
+        g_value_set_int(value, 0);
         break;
     case PROP_AUDIO:
         g_value_set_boolean(value, s->audio);
@@ -706,6 +711,9 @@ static void spice_session_get_property(GObject    *gobject,
     case PROP_PREF_COMPRESSION:
         g_value_set_enum(value, s->preferred_compression);
         break;
+    case PROP_GL_SCANOUT:
+        g_value_set_boolean(value, s->gl_scanout);
+        break;
     default:
 	G_OBJECT_WARN_INVALID_PROPERTY_ID(gobject, prop_id, pspec);
 	break;
@@ -814,7 +822,7 @@ static void spice_session_set_property(GObject      *gobject,
         s->secure_channels = g_value_dup_boxed(value);
         break;
     case PROP_COLOR_DEPTH:
-        s->color_depth = g_value_get_int(value);
+        spice_info("SpiceSession::color-depth has been deprecated. Property is ignored");
         break;
     case PROP_AUDIO:
         s->audio = g_value_get_boolean(value);
@@ -845,6 +853,13 @@ static void spice_session_set_property(GObject      *gobject,
     case PROP_PREF_COMPRESSION:
         s->preferred_compression = g_value_get_enum(value);
         break;
+    case PROP_GL_SCANOUT:
+#ifdef G_OS_UNIX
+        s->gl_scanout = g_value_get_boolean(value);
+#else
+        g_warning("SpiceSession:gl-scanout is only available on Unix");
+#endif
+        break;
     default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID(gobject, prop_id, pspec);
         break;
@@ -1112,6 +1127,9 @@ static void spice_session_class_init(SpiceSessionClass *klass)
      * Display color depth to set on new display channels. If 0, don't set.
      *
      * Since: 0.7
+     *
+     * Deprecated: 0.37: Deprecated due lack of support in drivers, only Windows 7 and older.
+     * This option is currently ignored.
      **/
     g_object_class_install_property
         (gobject_class, PROP_COLOR_DEPTH,
@@ -1119,6 +1137,7 @@ static void spice_session_class_init(SpiceSessionClass *klass)
                           "Color depth",
                           "Display channel color depth",
                           0, 32, 0,
+                          G_PARAM_DEPRECATED |
                           G_PARAM_READWRITE |
                           G_PARAM_STATIC_STRINGS));
 
@@ -1485,6 +1504,35 @@ static void spice_session_class_init(SpiceSessionClass *klass)
                            SPICE_IMAGE_COMPRESSION_INVALID,
                            G_PARAM_READWRITE |
                            G_PARAM_STATIC_STRINGS));
+
+    /**
+     * SpiceSession:gl-scanout:
+     *
+     * Whether to enable gl-scanout (Unix only).  Set to TRUE by
+     * default on EGL-enabled host, unless SPICE_DISABLE_GL_SCANOUT
+     * environment variable is set.
+     *
+     * Since: 0.36
+     **/
+    g_object_class_install_property
+        (gobject_class, PROP_GL_SCANOUT,
+         g_param_spec_boolean("gl-scanout",
+                              "Enable GL scanout support",
+                              "Enable GL scanout support",
+#ifdef HAVE_EGL
+                              g_getenv("SPICE_DISABLE_GL_SCANOUT") == NULL,
+                              G_PARAM_CONSTRUCT |
+#else
+                              false,
+#endif
+                              G_PARAM_READWRITE |
+                              G_PARAM_STATIC_STRINGS));
+}
+
+G_GNUC_INTERNAL
+gboolean spice_session_get_gl_scanout_enabled(SpiceSession *session)
+{
+    return session->priv->gl_scanout;
 }
 
 /* ------------------------------------------------------------------ */
@@ -1499,7 +1547,17 @@ static void spice_session_class_init(SpiceSessionClass *klass)
  **/
 SpiceSession *spice_session_new(void)
 {
-    return SPICE_SESSION(g_object_new(SPICE_TYPE_SESSION, NULL));
+    SpiceSession *self = SPICE_SESSION(g_object_new(SPICE_TYPE_SESSION, NULL));
+    SpiceSessionPrivate *priv = self->priv;
+    GError *err = NULL;
+
+    priv->usb_manager = spice_usb_device_manager_get(self, &err);
+    if (err != NULL) {
+        SPICE_DEBUG("Could not initialize SpiceUsbDeviceManager - %s", err->message);
+        g_clear_error(&err);
+    }
+
+    return self;
 }
 
 G_GNUC_INTERNAL
@@ -1663,23 +1721,21 @@ void spice_session_switching_disconnect(SpiceSession *self)
     g_return_if_fail(SPICE_IS_SESSION(self));
 
     SpiceSessionPrivate *s = self->priv;
-    struct channel *item;
-    RingItem *ring, *next;
 
     g_return_if_fail(s->cmain != NULL);
 
     /* disconnect/destroy all but main channel */
 
-    for (ring = ring_get_head(&s->channels); ring != NULL; ring = next) {
-        next = ring_next(&s->channels, ring);
-        item = SPICE_CONTAINEROF(ring, struct channel, link);
+    for (GList *l = s->channels; l != NULL; ) {
+        SpiceChannel *channel = l->data;
+        l = l->next;
 
-        if (item->channel == s->cmain)
+        if (channel == s->cmain)
             continue;
-        spice_session_channel_destroy(self, item->channel);
+        spice_session_channel_destroy(self, channel);
     }
 
-    g_warn_if_fail(!ring_is_empty(&s->channels)); /* ring_get_length() == 1 */
+    g_warn_if_fail(s->channels != NULL);
 
     cache_clear_all(self);
     s->connection_id = 0;
@@ -1717,10 +1773,10 @@ void spice_session_start_migrating(SpiceSession *session,
     SWAP_STR(s->tls_port, m->tls_port);
     SWAP_STR(s->unix_path, m->unix_path);
 
-    g_warn_if_fail(ring_get_length(&s->channels) == ring_get_length(&m->channels));
+    g_warn_if_fail(g_list_length(s->channels) == g_list_length(m->channels));
 
     SPICE_DEBUG("migration channels left:%u (in migration:%u)",
-                ring_get_length(&s->channels), ring_get_length(&m->channels));
+                g_list_length(s->channels), g_list_length(m->channels));
     s->migration_left = spice_session_get_channels(session);
 }
 #undef SWAP_STR
@@ -1730,26 +1786,21 @@ SpiceChannel* spice_session_lookup_channel(SpiceSession *session, gint id, gint
 {
     g_return_val_if_fail(SPICE_IS_SESSION(session), NULL);
 
-    RingItem *ring, *next;
     SpiceSessionPrivate *s = session->priv;
-    struct channel *c;
+    SpiceChannel *channel = NULL;
 
-    for (ring = ring_get_head(&s->channels);
-         ring != NULL; ring = next) {
-        next = ring_next(&s->channels, ring);
-        c = SPICE_CONTAINEROF(ring, struct channel, link);
-        if (c == NULL || c->channel == NULL) {
-            g_warn_if_reached();
-            continue;
-        }
+    for (GList *l = s->channels; l != NULL; ) {
+        channel = l->data;
+        l = l->next;
 
-        if (id == spice_channel_get_channel_id(c->channel) &&
-            type == spice_channel_get_channel_type(c->channel))
+        if (id == spice_channel_get_channel_id(channel) &&
+            type == spice_channel_get_channel_type(channel)) {
             break;
+        }
     }
-    g_return_val_if_fail(ring != NULL, NULL);
+    g_return_val_if_fail(channel != NULL, NULL);
 
-    return c->channel;
+    return channel;
 }
 
 G_GNUC_INTERNAL
@@ -1758,8 +1809,6 @@ void spice_session_abort_migration(SpiceSession *session)
     g_return_if_fail(SPICE_IS_SESSION(session));
 
     SpiceSessionPrivate *s = session->priv;
-    RingItem *ring, *next;
-    struct channel *c;
 
     if (s->migration == NULL) {
         SPICE_DEBUG("no migration in progress");
@@ -1770,18 +1819,17 @@ void spice_session_abort_migration(SpiceSession *session)
     if (s->migration_state != SPICE_SESSION_MIGRATION_MIGRATING)
         goto end;
 
-    for (ring = ring_get_head(&s->channels);
-         ring != NULL; ring = next) {
-        next = ring_next(&s->channels, ring);
-        c = SPICE_CONTAINEROF(ring, struct channel, link);
+    for (GList *l = s->channels; l != NULL; ) {
+        SpiceChannel *channel = l->data;
+        l = l->next;
 
-        if (g_list_find(s->migration_left, c->channel))
+        if (g_list_find(s->migration_left, channel))
             continue;
 
-        spice_channel_swap(c->channel,
+        spice_channel_swap(channel,
             spice_session_lookup_channel(s->migration,
-                                         spice_channel_get_channel_id(c->channel),
-                                         spice_channel_get_channel_type(c->channel)),
+                                         spice_channel_get_channel_id(channel),
+                                         spice_channel_get_channel_type(channel)),
                                          !s->full_migration);
     }
 
@@ -1880,14 +1928,13 @@ void spice_session_migrate_end(SpiceSession *self)
 
     SpiceSessionPrivate *s = self->priv;
     SpiceMsgOut *out;
-    GList *l;
 
     g_return_if_fail(s->migration);
     g_return_if_fail(s->migration->priv->cmain);
     g_return_if_fail(g_list_length(s->migration_left) != 0);
 
     /* disconnect and reset all channels */
-    for (l = s->migration_left; l != NULL; ) {
+    for (GList *l = s->migration_left; l != NULL; ) {
         SpiceChannel *channel = l->data;
         l = l->next;
 
@@ -1975,23 +2022,10 @@ void spice_session_disconnect(SpiceSession *session)
  **/
 GList *spice_session_get_channels(SpiceSession *session)
 {
-    SpiceSessionPrivate *s;
-    struct channel *item;
-    GList *list = NULL;
-    RingItem *ring;
-
     g_return_val_if_fail(SPICE_IS_SESSION(session), NULL);
     g_return_val_if_fail(session->priv != NULL, NULL);
 
-    s = session->priv;
-
-    for (ring = ring_get_head(&s->channels);
-         ring != NULL;
-         ring = ring_next(&s->channels, ring)) {
-        item = SPICE_CONTAINEROF(ring, struct channel, link);
-        list = g_list_append(list, item->channel);
-    }
-    return list;
+    return g_list_copy(session->priv->channels);
 }
 
 /**
@@ -2007,19 +2041,15 @@ GList *spice_session_get_channels(SpiceSession *session)
 gboolean spice_session_has_channel_type(SpiceSession *session, gint type)
 {
     SpiceSessionPrivate *s;
-    struct channel *item;
-    RingItem *ring;
 
     g_return_val_if_fail(SPICE_IS_SESSION(session), FALSE);
     g_return_val_if_fail(session->priv != NULL, FALSE);
 
     s = session->priv;
 
-    for (ring = ring_get_head(&s->channels);
-         ring != NULL;
-         ring = ring_next(&s->channels, ring)) {
-        item = SPICE_CONTAINEROF(ring, struct channel, link);
-        if (spice_channel_get_channel_type(item->channel) == type) {
+    for (GList *l = s->channels; l != NULL; l = l->next) {
+        SpiceChannel *channel = l->data;
+        if (spice_channel_get_channel_type(channel) == type) {
             return TRUE;
         }
     }
@@ -2233,6 +2263,23 @@ GSocketConnection* spice_session_channel_open_host(SpiceSession *session, SpiceC
         g_socket_set_timeout(socket, 0);
         g_socket_set_blocking(socket, FALSE);
         g_socket_set_keepalive(socket, TRUE);
+
+        /* Make client timeouts a bit more responsive */
+#if defined(_WIN32)
+        /*  Windows does not support setting count */
+        struct tcp_keepalive keepalive = {
+            TRUE,
+            30 * 1000,
+            5 * 1000
+        };
+        DWORD written;
+        WSAIoctl(g_socket_get_fd(socket), SIO_KEEPALIVE_VALS, &keepalive, sizeof(keepalive),
+                 NULL, 0, &written, NULL, NULL);
+#elif defined(TCP_KEEPIDLE) && defined(TCP_KEEPINTVL)
+        g_socket_set_option(socket, SOL_TCP, TCP_KEEPIDLE, 30, NULL);
+        g_socket_set_option(socket, SOL_TCP, TCP_KEEPINTVL, 15, NULL);
+        g_socket_set_option(socket, SOL_TCP, TCP_KEEPCNT, 3, NULL);
+#endif
     }
 
     g_clear_object(&open_host.client);
@@ -2247,12 +2294,8 @@ void spice_session_channel_new(SpiceSession *session, SpiceChannel *channel)
     g_return_if_fail(SPICE_IS_CHANNEL(channel));
 
     SpiceSessionPrivate *s = session->priv;
-    struct channel *item;
 
-
-    item = g_new0(struct channel, 1);
-    item->channel = channel;
-    ring_add(&s->channels, &item->link);
+    s->channels = g_list_prepend(s->channels, channel);
 
     if (SPICE_IS_MAIN_CHANNEL(channel)) {
         gboolean all = spice_strv_contains(s->disable_effects, "all");
@@ -2262,8 +2305,6 @@ void spice_session_channel_new(SpiceSession *session, SpiceChannel *channel)
                      "disable-font-smooth", all || spice_strv_contains(s->disable_effects, "font-smooth"),
                      "disable-animation", all || spice_strv_contains(s->disable_effects, "animation"),
                      NULL);
-        if (s->color_depth != 0)
-            g_object_set(channel, "color-depth", s->color_depth, NULL);
 
         CHANNEL_DEBUG(channel, "new main channel, switching");
         s->cmain = channel;
@@ -2280,7 +2321,7 @@ static void channel_finally_destroyed(gpointer data, GObject *channel)
     SpiceSession *session = SPICE_SESSION(data);
     SpiceSessionPrivate *s = session->priv;
     s->channels_destroying--;
-    if (ring_is_empty(&s->channels) && (s->channels_destroying == 0)) {
+    if (s->channels == NULL && (s->channels_destroying == 0)) {
         g_signal_emit(session, signals[SPICE_SESSION_DISCONNECTED], 0);
     }
     g_object_unref(session);
@@ -2292,28 +2333,24 @@ static void spice_session_channel_destroy(SpiceSession *session, SpiceChannel *c
     g_return_if_fail(SPICE_IS_CHANNEL(channel));
 
     SpiceSessionPrivate *s = session->priv;
-    struct channel *item = NULL;
-    RingItem *ring;
+    GList *l;
 
     if (s->migration_left)
         s->migration_left = g_list_remove(s->migration_left, channel);
 
-    for (ring = ring_get_head(&s->channels); ring != NULL;
-         ring = ring_next(&s->channels, ring)) {
-        item = SPICE_CONTAINEROF(ring, struct channel, link);
-        if (item->channel == channel)
+    for (l = s->channels; l != NULL; l = l->next) {
+        if (l->data == channel)
             break;
     }
 
-    g_return_if_fail(ring != NULL);
+    g_return_if_fail(l != NULL);
 
     if (channel == s->cmain) {
         CHANNEL_DEBUG(channel, "the session lost the main channel");
         s->cmain = NULL;
     }
 
-    ring_remove(&item->link);
-    g_free(item);
+    s->channels = g_list_delete_link(s->channels, l);
 
     g_signal_emit(session, signals[SPICE_SESSION_CHANNEL_DESTROY], 0, channel);
 
@@ -2357,7 +2394,7 @@ guint32 spice_session_get_mm_time(SpiceSession *session)
 
     /* FIXME: we may want to estimate the drift of clocks, and well,
        do something better than this trivial approach */
-    return s->mm_time + (g_get_monotonic_time() - s->mm_time_at_clock) / 1000;
+    return (g_get_monotonic_time() - s->mm_time_offset) / 1000;
 }
 
 #define MM_TIME_DIFF_RESET_THRESH 500 // 0.5 sec
@@ -2372,12 +2409,11 @@ void spice_session_set_mm_time(SpiceSession *session, guint32 time)
 
     old_time = spice_session_get_mm_time(session);
 
-    s->mm_time = time;
-    s->mm_time_at_clock = g_get_monotonic_time();
-    SPICE_DEBUG("set mm time: %u", spice_session_get_mm_time(session));
+    s->mm_time_offset = g_get_monotonic_time() - time * (guint64) 1000;
+    SPICE_DEBUG("set mm time: %u", time);
     if (spice_mmtime_diff(time, old_time + MM_TIME_DIFF_RESET_THRESH) > 0 ||
         spice_mmtime_diff(time, old_time) < 0) {
-        SPICE_DEBUG("%s: mm-time-reset, old %u, new %u", __FUNCTION__, old_time, s->mm_time);
+        SPICE_DEBUG("%s: mm-time-reset, old %u, new %u", __FUNCTION__, old_time, time);
         g_coroutine_signal_emit(session, signals[SPICE_SESSION_MM_TIME_RESET], 0);
     }
 }
@@ -2620,8 +2656,7 @@ guint32 spice_session_get_playback_latency(SpiceSession *session)
     }
 }
 
-G_GNUC_INTERNAL
-const gchar* spice_session_get_shared_dir(SpiceSession *session)
+static const gchar* spice_session_get_shared_dir(SpiceSession *session)
 {
     g_return_val_if_fail(SPICE_IS_SESSION(session), NULL);
 
@@ -2630,8 +2665,7 @@ const gchar* spice_session_get_shared_dir(SpiceSession *session)
     return s->shared_dir;
 }
 
-G_GNUC_INTERNAL
-void spice_session_set_shared_dir(SpiceSession *session, const gchar *dir)
+static void spice_session_set_shared_dir(SpiceSession *session, const gchar *dir)
 {
     g_return_if_fail(SPICE_IS_SESSION(session));
 
@@ -2823,16 +2857,6 @@ gboolean spice_session_is_for_migration(SpiceSession *session)
     return session->priv->for_migration;
 }
 
-G_GNUC_INTERNAL
-void spice_session_set_main_channel(SpiceSession *session, SpiceChannel *channel)
-{
-    g_return_if_fail(SPICE_IS_SESSION(session));
-    g_return_if_fail(SPICE_IS_CHANNEL(channel));
-    g_return_if_fail(session->priv->cmain == NULL);
-
-    session->priv->cmain = channel;
-}
-
 G_GNUC_INTERNAL
 gboolean spice_session_set_migration_session(SpiceSession *session, SpiceSession *mig_session)
 {
