@@ -24,10 +24,6 @@
 #include <termios.h>
 #endif
 
-#ifdef USE_SMARTCARD_012
-#include <vreader.h>
-#endif
-
 #include "spice-widget.h"
 #include "spice-gtk-session.h"
 #include "spice-audio.h"
@@ -37,9 +33,7 @@
 #include "usb-device-widget.h"
 
 #include "spicy-connect.h"
-#if HAVE_GSTAUDIO || HAVE_GSTVIDEO
 #include <gst/gst.h>
-#endif
 
 typedef struct spice_connection spice_connection;
 
@@ -89,7 +83,7 @@ struct _SpiceWindowClass
 
 static GType spice_window_get_type(void);
 
-G_DEFINE_TYPE (SpiceWindow, spice_window, G_TYPE_OBJECT);
+G_DEFINE_TYPE (SpiceWindow, spice_window, G_TYPE_OBJECT)
 
 #define CHANNELID_MAX 4
 #define MONITORID_MAX 4
@@ -106,7 +100,6 @@ struct spice_connection {
     const char       *mouse_state;
     const char       *agent_state;
     gboolean         agent_connected;
-    int              channels;
     int              disconnecting;
 
     /* key: SpiceFileTransferTask, value: TransferTaskWidgets */
@@ -117,7 +110,8 @@ struct spice_connection {
 static spice_connection *connection_new(void);
 static void connection_connect(spice_connection *conn);
 static void connection_disconnect(spice_connection *conn);
-static void connection_destroy(spice_connection *conn);
+static void connection_destroy(SpiceSession *session,
+                               spice_connection *conn);
 static void usb_connect_failed(GObject               *object,
                                SpiceUsbDevice        *device,
                                GError                *error,
@@ -1732,7 +1726,6 @@ static void channel_new(SpiceSession *s, SpiceChannel *channel, gpointer data)
     int id;
 
     g_object_get(channel, "channel-id", &id, NULL);
-    conn->channels++;
     SPICE_DEBUG("new channel (#%d)", id);
 
     if (SPICE_IS_MAIN_CHANNEL(channel)) {
@@ -1815,13 +1808,6 @@ static void channel_destroy(SpiceSession *s, SpiceChannel *channel, gpointer dat
         if (SPICE_PORT_CHANNEL(channel) == stdin_port)
             stdin_port = NULL;
     }
-
-    conn->channels--;
-    if (conn->channels > 0) {
-        return;
-    }
-
-    connection_destroy(conn);
 }
 
 static void migration_state(GObject *session,
@@ -1848,6 +1834,8 @@ static spice_connection *connection_new(void)
                      G_CALLBACK(channel_destroy), conn);
     g_signal_connect(conn->session, "notify::migration-state",
                      G_CALLBACK(migration_state), conn);
+    g_signal_connect(conn->session, "disconnected",
+                     G_CALLBACK(connection_destroy), conn);
 
     manager = spice_usb_device_manager_get(conn->session, NULL);
     if (manager) {
@@ -1879,7 +1867,8 @@ static void connection_disconnect(spice_connection *conn)
     spice_session_disconnect(conn->session);
 }
 
-static void connection_destroy(spice_connection *conn)
+static void connection_destroy(SpiceSession *session,
+                               spice_connection *conn)
 {
     g_object_unref(conn->session);
     g_hash_table_unref(conn->transfers);
@@ -2001,9 +1990,7 @@ int main(int argc, char *argv[])
 
     /* parse opts */
     gtk_init(&argc, &argv);
-#if HAVE_GSTAUDIO || HAVE_GSTVIDEO
     gst_init(&argc, &argv);
-#endif
     context = g_option_context_new("- spice client test application");
     g_option_context_set_summary(context, "Gtk+ test client to connect to Spice servers.");
     g_option_context_set_description(context, "Report bugs to " PACKAGE_BUGREPORT ".");
@@ -2011,9 +1998,7 @@ int main(int argc, char *argv[])
     g_option_context_set_main_group(context, spice_cmdline_get_option_group());
     g_option_context_add_main_entries(context, cmd_entries, NULL);
     g_option_context_add_group(context, gtk_get_option_group(TRUE));
-#if HAVE_GSTAUDIO || HAVE_GSTVIDEO
     g_option_context_add_group(context, gst_init_get_option_group());
-#endif
     if (!g_option_context_parse (context, &argc, &argv, &error)) {
         g_print("option parsing failed: %s\n", error->message);
         exit(1);
@@ -2068,8 +2053,6 @@ int main(int argc, char *argv[])
     g_free(spicy_title);
 
     setup_terminal(true);
-#if HAVE_GSTAUDIO || HAVE_GSTVIDEO
     gst_deinit();
-#endif
     return 0;
 }
